#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
This module contains the class and auxiliary methods of a model.
"""
import logging
import os
import os.path
import signal
import sys
import threading
import time
from abc import ABC, abstractmethod
from collections import OrderedDict

import tensorflow as tf
from tabulate import tabulate
from tqdm import tqdm

from ludwig.constants import LOSS, COMBINED, TRAINING, VALIDATION, TEST, TYPE
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.globals import MODEL_WEIGHTS_FILE_NAME
from ludwig.globals import TRAINING_CHECKPOINTS_DIR_PATH
from ludwig.globals import TRAINING_PROGRESS_TRACKER_FILE_NAME
from ludwig.globals import is_progressbar_disabled
from ludwig.models.predictor import Predictor
from ludwig.modules.metric_modules import get_improved_fun
from ludwig.modules.metric_modules import get_initial_validation_value
from ludwig.modules.optimization_modules import ClippedOptimizer
from ludwig.utils import time_utils
from ludwig.utils.data_utils import load_json, save_json
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.horovod_utils import initialize_horovod, return_first
from ludwig.utils.math_utils import learning_rate_warmup, \
    learning_rate_warmup_distributed, exponential_decay
from ludwig.utils.misc_utils import set_random_seed
from ludwig.utils.tf_utils import initialize_tensorflow

logger = logging.getLogger(__name__)


class BaseTrainer(ABC):
    @abstractmethod
    def train(
            self,
            model,
            training_set,
            validation_set=None,
            test_set=None,
            save_path='model',
            **kwargs
    ):
        raise NotImplementedError()

    @abstractmethod
    def train_online(
            self,
            model,
            dataset,
    ):
        raise NotImplementedError()

    @property
    @abstractmethod
    def validation_field(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def validation_metric(self):
        raise NotImplementedError()

    # Remote implementations may override this
    def shutdown(self):
        pass

    # Functions needed to treat Trainer as a context manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()


class Trainer(BaseTrainer):
    """
    Trainer is a class that train a model
    """

    def __init__(
            self,
            optimizer=None,
            epochs=100,
            regularization_lambda=0.0,
            learning_rate=0.001,
            decay=False,
            decay_rate=0.96,
            decay_steps=10000,
            staircase=False,
            batch_size=128,
            eval_batch_size=0,
            should_shuffle=True,
            shuffle_buffer_size=None,
            bucketing_field=None,
            validation_field='combined',
            validation_metric='loss',
            early_stop=20,
            reduce_learning_rate_on_plateau=0,
            reduce_learning_rate_on_plateau_patience=5,
            reduce_learning_rate_on_plateau_rate=0.5,
            reduce_learning_rate_eval_metric=LOSS,
            reduce_learning_rate_eval_split=TRAINING,
            increase_batch_size_on_plateau=0,
            increase_batch_size_on_plateau_patience=5,
            increase_batch_size_on_plateau_rate=2,
            increase_batch_size_on_plateau_max=512,
            increase_batch_size_eval_metric=LOSS,
            increase_batch_size_eval_split=TRAINING,
            learning_rate_warmup_epochs=1,
            resume=False,
            skip_save_model=False,
            skip_save_progress=False,
            skip_save_log=False,
            callbacks=None,
            random_seed=default_random_seed,
            horovod=None,
            debug=False,
            **kwargs
    ):
        """Trains a model with a set of hyperparameters listed below. Customizable
                :param training_set: The training set
        :param validation_set: The validation dataset
        :param test_set: The test dataset
        :param validation_field: The first output feature, by default it is set
               as the same field of the first output feature.
        :param validation_metric: metric used on the validation field, it is
               accuracy by default
        :type validation_metric:
        :param save_path: The path to save the file
        :type save_path: filepath (str)
        :param regularization_lambda: Strength of the $L2$ regularization
        :type regularization_lambda: Integer
        :param epochs: Number of epochs the algorithm is intended to be run over
        :type epochs: Integer
        :param learning_rate: Learning rate for the algorithm, represents how
               much to scale the gradients by
        :type learning_rate: Integer
        :param batch_size: Size of batch to pass to the model for training.
        :type batch_size: Integer
        :param eval_batch_size: Size of batch to pass to the model for evaluation.
        :type eval_batch_size: Integer
        :param should_shuffle: Shuffle batches during training when true (default: True).
        :type shuffle_buffer_size: Boolean
        :param shuffle_buffer_size: Size of buffer in number of examples to read for shuffling.
        :type shuffle_buffer_size: Integer
        :param bucketing_field: when batching, buckets datapoints based the
               length of a field together. Bucketing on text length speeds up
               training of RNNs consistently, 30% in some cases
        :type bucketing_field:
        :param validation_field: The first output feature, by default it is set
               as the same field of the first output feature.
        :param validation_metric: metric used on the validation field, it is
               accuracy by default
        :type validation_metric:
        :param dropout: dropout probability (probability of dropping
               a neuron in a given layer)
        :type dropout: Float
        :param early_stop: How many epochs without any improvement in the
               validation_metric triggers the algorithm to stop
        :type early_stop: Integer
        :param reduce_learning_rate_on_plateau: Reduces the learning rate when
               the algorithm hits a plateau (i.e. the performance on the
               validation does not improve)
        :type reduce_learning_rate_on_plateau: Float
        :param reduce_learning_rate_on_plateau_patience: How many epochs have
               to pass before the learning rate reduces
        :type reduce_learning_rate_on_plateau_patience: Float
        :param reduce_learning_rate_on_plateau_rate: Rate at which we reduce
               the learning rate
        :type reduce_learning_rate_on_plateau_rate: Float
        :param increase_batch_size_on_plateau: Increase the batch size on a
               plateau
        :type increase_batch_size_on_plateau: Integer
        :param increase_batch_size_on_plateau_patience: How many epochs to wait
               for before increasing the batch size
        :type increase_batch_size_on_plateau_patience: Integer
        :param increase_batch_size_on_plateau_rate: The rate at which the batch
               size increases.
        :type increase_batch_size_on_plateau_rate: Float
        :param increase_batch_size_on_plateau_max: The maximum size of the batch
        :type increase_batch_size_on_plateau_max: Integer
        :param learning_rate_warmup_epochs: The number of epochs to warmup the
               learning rate for.
        :type learning_rate_warmup_epochs: Integer
        :param resume: Resume training a model that was being trained.
        :type resume: Boolean
        :param skip_save_model: disables
               saving model weights and hyperparameters each time the model
               improves. By default Ludwig saves model weights after each epoch
               the validation metric (improves, but if the model is really big
               that can be time consuming. If you do not want to keep
               the weights and just find out what performance a model can get
               with a set of hyperparameters, use this parameter to skip it,
               but the model will not be loadable later on.
        :type skip_save_model: Boolean
        :param skip_save_progress: disables saving progress each epoch.
               By default Ludwig saves weights and stats  after each epoch
               for enabling resuming of training, but if the model is
               really big that can be time consuming and will uses twice
               as much space, use this parameter to skip it, but training
               cannot be resumed later on
        :type skip_save_progress: Boolean
        :param skip_save_log: Disables saving TensorBoard
               logs. By default Ludwig saves logs for the TensorBoard, but if it
               is not needed turning it off can slightly increase the
               overall speed..
        :type skip_save_log: Boolean
        :param callbacks: a list of `ludwig.callbacks.Callback` objects that
               provide hooks into the Ludwig pipeline.
        :type: list
        :param random_seed: Default initialization for the random seeds
        :type: Float
        """
        self.epochs = epochs
        self.regularization_lambda = regularization_lambda
        self.learning_rate = learning_rate
        self.decay = decay
        self.decay_rate = decay_rate
        self.decay_steps = decay_steps
        self.staircase = staircase
        self.batch_size = batch_size
        self.eval_batch_size = batch_size if eval_batch_size < 1 else eval_batch_size
        self.should_shuffle = should_shuffle
        self.shuffle_buffer_size = shuffle_buffer_size
        self.bucketing_field = bucketing_field
        self._validation_field = validation_field
        self._validation_metric = validation_metric
        self.early_stop = early_stop
        self.reduce_learning_rate_on_plateau = reduce_learning_rate_on_plateau
        self.reduce_learning_rate_on_plateau_patience = reduce_learning_rate_on_plateau_patience
        self.reduce_learning_rate_on_plateau_rate = reduce_learning_rate_on_plateau_rate
        self.reduce_learning_rate_eval_metric = reduce_learning_rate_eval_metric
        self.reduce_learning_rate_eval_split = reduce_learning_rate_eval_split
        self.increase_batch_size_on_plateau = increase_batch_size_on_plateau
        self.increase_batch_size_on_plateau_patience = increase_batch_size_on_plateau_patience
        self.increase_batch_size_on_plateau_rate = increase_batch_size_on_plateau_rate
        self.increase_batch_size_on_plateau_max = increase_batch_size_on_plateau_max
        self.increase_batch_size_eval_metric = increase_batch_size_eval_metric
        self.increase_batch_size_eval_split = increase_batch_size_eval_split
        self.learning_rate_warmup_epochs = learning_rate_warmup_epochs
        self.resume = resume
        self.skip_save_model = skip_save_model
        self.skip_save_progress = skip_save_progress
        self.skip_save_log = skip_save_log
        self.random_seed = random_seed
        self.horovod = horovod
        self.debug = debug
        self.received_sigint = False
        self.callbacks = callbacks or []

        if self.horovod:
            self.learning_rate *= self.horovod.size()

        # ================ Optimizer ================
        if optimizer is None:
            optimizer = {TYPE: 'Adam'}
        self.optimizer = ClippedOptimizer(
            horovod=horovod,
            **optimizer
        )

    @classmethod
    def write_epoch_summary(
            cls,
            summary_writer,
            metrics,
            step,
    ):
        if not summary_writer:
            return

        with summary_writer.as_default():
            for feature_name, output_feature in metrics.items():
                for metric in output_feature:
                    metric_tag = "{}/epoch_{}".format(
                        feature_name, metric
                    )
                    metric_val = output_feature[metric][-1]
                    tf.summary.scalar(metric_tag, metric_val, step=step)
        summary_writer.flush()

    @classmethod
    def write_step_summary(
            cls,
            train_summary_writer,
            combined_loss,
            all_losses,
            step,
            learning_rate=None
    ):
        if not train_summary_writer:
            return

        with train_summary_writer.as_default():
            # combined loss
            loss_tag = "{}/step_training_loss".format("combined")
            tf.summary.scalar(loss_tag, combined_loss, step=step)

            # all other losses
            for feature_name, loss in all_losses.items():
                loss_tag = "{}/step_training_loss".format(feature_name)
                tf.summary.scalar(loss_tag, loss, step=step)

            if learning_rate:
                tf.summary.scalar("combined/step_learning_rate",
                                  learning_rate, step=step)

        train_summary_writer.flush()

    def train(
            self,
            model,
            training_set,
            validation_set=None,
            test_set=None,
            save_path='model',
            **kwargs
    ):
        """Trains a model with a set of hyperparameters listed below. Customizable
        :param training_set: The training set
        :param validation_set: The validation dataset
        :param test_set: The test dataset
        """
        # ====== General setup =======
        output_features = model.output_features
        digits_per_epochs = len(str(self.epochs))
        # Only use signals when on the main thread to avoid issues with CherryPy: https://github.com/ludwig-ai/ludwig/issues/286
        if threading.current_thread() == threading.main_thread():
            signal.signal(signal.SIGINT, self.set_epochs_to_1_or_quit)
        should_validate = validation_set is not None and validation_set.size > 0

        metrics_names = self.get_metrics_names(output_features)

        # check if validation_field is valid
        valid_validation_field = False
        if self.validation_field == 'combined':
            valid_validation_field = True
            if self.validation_metric is not LOSS and len(
                    output_features) == 1:
                only_of = next(iter(output_features))
                if self.validation_metric in metrics_names[only_of]:
                    self._validation_field = only_of
                    logger.warning(
                        "Replacing 'combined' validation field "
                        "with '{}' as the specified validation "
                        "metric {} is invalid for 'combined' "
                        "but is valid for '{}'.".format(
                            only_of, self.validation_metric, only_of
                        ))
        else:
            for output_feature in output_features:
                if self.validation_field == output_feature:
                    valid_validation_field = True

        if not valid_validation_field:
            raise ValueError(
                'The specified validation_field {} is not valid.'
                'Available ones are: {}'.format(
                    self.validation_field,
                    list(output_features.keys()) + ['combined']
                )
            )

        # check if validation_metric is valid
        valid_validation_metric = self.validation_metric in metrics_names[
            self.validation_field
        ]
        if not valid_validation_metric:
            raise ValueError(
                'The specified metric {} is not valid. '
                'Available metrics for {} output feature are: {}'.format(
                    self.validation_metric,
                    self.validation_field,
                    metrics_names[self.validation_field]
                )
            )

        # ====== Setup file names =======
        model_weights_path = model_hyperparameters_path = None
        training_checkpoints_path = training_checkpoints_prefix_path = training_progress_tracker_path = None
        tensorboard_log_dir = None
        if self.is_coordinator():
            os.makedirs(save_path, exist_ok=True)
            model_weights_path = os.path.join(save_path,
                                              MODEL_WEIGHTS_FILE_NAME)
            model_hyperparameters_path = os.path.join(
                save_path, MODEL_HYPERPARAMETERS_FILE_NAME
            )
            training_checkpoints_path = os.path.join(
                save_path, TRAINING_CHECKPOINTS_DIR_PATH
            )
            # training_checkpoints_prefix_path = os.path.join(
            #    training_checkpoints_path, "ckpt"
            # )
            training_progress_tracker_path = os.path.join(
                save_path, TRAINING_PROGRESS_TRACKER_FILE_NAME
            )
            tensorboard_log_dir = os.path.join(
                save_path, 'logs'
            )

        # ====== Setup session =======
        checkpoint = checkpoint_manager = None
        if self.is_coordinator():
            checkpoint = tf.train.Checkpoint(
                optimizer=self.optimizer,
                model=model
            )
            checkpoint_manager = tf.train.CheckpointManager(
                checkpoint, training_checkpoints_path, max_to_keep=1
            )

        train_summary_writer = None
        validation_summary_writer = None
        test_summary_writer = None
        if self.is_coordinator() and not self.skip_save_log and tensorboard_log_dir:
            train_summary_writer = tf.summary.create_file_writer(
                os.path.join(
                    tensorboard_log_dir, TRAINING
                )
            )
            if validation_set is not None and validation_set.size > 0:
                validation_summary_writer = tf.summary.create_file_writer(
                    os.path.join(
                        tensorboard_log_dir, VALIDATION
                    )
                )
            if test_set is not None and test_set.size > 0:
                test_summary_writer = tf.summary.create_file_writer(
                    os.path.join(
                        tensorboard_log_dir, TEST
                    )
                )

        if self.debug and self.is_coordinator():
            # See https://www.tensorflow.org/tensorboard/debugger_v2 for usage.
            # debug_path = os.path.join(
            #     save_path, 'debug'
            # )
            # tf.debugging.experimental.enable_dump_debug_info(
            #     debug_path,
            #     tensor_debug_mode='FULL_HEALTH',
            #     circular_buffer_size=-1,
            # )
            tf.config.experimental_run_functions_eagerly(True)
            # tf.debugging.enable_check_numerics()

        # ================ Resume logic ================
        if self.resume:
            progress_tracker = self.resume_training_progress_tracker(
                training_progress_tracker_path
            )
            if self.is_coordinator():
                self.resume_weights_and_optimzier(
                    training_checkpoints_path, checkpoint
                )
        else:
            (
                train_metrics,
                vali_metrics,
                test_metrics
            ) = self.initialize_training_metrics(output_features)

            progress_tracker = ProgressTracker(
                batch_size=self.batch_size,
                epoch=0,
                steps=0,
                last_improvement_epoch=0,
                last_learning_rate_reduction_epoch=0,
                last_increase_batch_size_epoch=0,
                learning_rate=self.learning_rate,
                best_eval_metric=get_initial_validation_value(
                    self.validation_metric
                ),
                best_reduce_learning_rate_eval_metric=get_initial_validation_value(
                    self.reduce_learning_rate_eval_metric
                ),
                last_reduce_learning_rate_eval_metric_improvement=0,
                best_increase_batch_size_eval_metric=get_initial_validation_value(
                    self.increase_batch_size_eval_metric
                ),
                last_increase_batch_size_eval_metric_improvement=0,
                num_reductions_learning_rate=0,
                num_increases_batch_size=0,
                train_metrics=train_metrics,
                vali_metrics=vali_metrics,
                test_metrics=test_metrics,
                last_improvement=0,
                last_learning_rate_reduction=0,
                last_increase_batch_size=0,
            )

        set_random_seed(self.random_seed)
        with training_set.initialize_batcher(
            batch_size=self.batch_size,
            should_shuffle=self.should_shuffle,
            shuffle_buffer_size=self.shuffle_buffer_size,
            seed=self.random_seed,
            horovod=self.horovod,
        ) as batcher:

            # ================ Training Loop ================
            first_batch = True
            while progress_tracker.epoch < self.epochs:
                batcher.set_epoch(progress_tracker.epoch)

                # epoch init
                start_time = time.time()
                if self.is_coordinator():
                    logger.info(
                        '\nEpoch {epoch:{digits}d}'.format(
                            epoch=progress_tracker.epoch + 1,
                            digits=digits_per_epochs
                        )
                    )

                # needed because batch size may change
                batcher.batch_size = progress_tracker.batch_size

                # Reset the metrics at the start of the next epoch
                model.reset_metrics()

                # ================ Train ================
                progress_bar = None
                if self.is_coordinator():
                    progress_bar = tqdm(
                        desc='Training',
                        total=batcher.steps_per_epoch,
                        file=sys.stdout,
                        disable=is_progressbar_disabled()
                    )

                self.callback(lambda c: c.on_epoch_start(self, progress_tracker, save_path))

                # training step loop
                while not batcher.last_batch():
                    self.callback(lambda c: c.on_batch_start(self, progress_tracker, save_path))

                    # Set learning rate for this batch
                    current_learning_rate = progress_tracker.learning_rate

                    if self.decay:
                        current_learning_rate = exponential_decay(
                            current_learning_rate,
                            self.decay_rate,
                            self.decay_steps,
                            progress_tracker.steps,
                            self.staircase
                        )

                    if self.horovod:
                        current_learning_rate = learning_rate_warmup_distributed(
                            current_learning_rate,
                            progress_tracker.epoch,
                            self.learning_rate_warmup_epochs,
                            self.horovod.size(),
                            batcher.step,
                            batcher.steps_per_epoch
                        ) * self.horovod.size()
                    else:
                        current_learning_rate = learning_rate_warmup(
                            current_learning_rate,
                            progress_tracker.epoch,
                            self.learning_rate_warmup_epochs,
                            batcher.step,
                            batcher.steps_per_epoch
                        )
                    self.optimizer.set_learning_rate(current_learning_rate)

                    # obtain batch
                    batch = batcher.next_batch()
                    inputs = {
                        i_feat.feature_name: batch[i_feat.proc_column]
                        for i_feat in model.input_features.values()
                    }
                    targets = {
                        o_feat.feature_name: batch[o_feat.proc_column]
                        for o_feat in model.output_features.values()
                    }

                    # Reintroduce for tensorboard graph
                    # if first_batch and self.is_coordinator() and not skip_save_log:
                    #    tf.summary.trace_on(graph=True, profiler=True)

                    loss, all_losses = model.train_step(
                        self.optimizer,
                        inputs,
                        targets,
                        self.regularization_lambda
                    )

                    # Reintroduce for tensorboard graph
                    # if first_batch and self.is_coordinator() and not skip_save_log:
                    #     with train_summary_writer.as_default():
                    #         tf.summary.trace_export(
                    #             name="Model",
                    #             step=0,
                    #             profiler_outdir=tensorboard_log_dir
                    #         )

                    if self.is_coordinator() and not self.skip_save_log:
                        self.write_step_summary(
                            train_summary_writer=train_summary_writer,
                            combined_loss=loss,
                            all_losses=all_losses,
                            step=progress_tracker.steps,
                            learning_rate=current_learning_rate,
                        )

                    if self.horovod and first_batch:
                        # Horovod: broadcast initial variable states from rank 0 to all other processes.
                        # This is necessary to ensure consistent initialization of all workers when
                        # training is started with random weights or restored from a checkpoint.
                        #
                        # Note: broadcast should be done after the first gradient step to ensure
                        # optimizer initialization.
                        self.horovod.broadcast_variables(model.variables,
                                                         root_rank=0)
                        self.horovod.broadcast_variables(
                            self.optimizer.variables(), root_rank=0)

                    progress_tracker.steps += 1
                    if self.is_coordinator():
                        progress_bar.update(1)
                    first_batch = False

                    self.callback(lambda c: c.on_batch_end(self, progress_tracker, save_path))

                # ================ Post Training Epoch ================
                if self.is_coordinator():
                    progress_bar.close()

                progress_tracker.epoch += 1

                # ================ Eval ================
                # init tables
                tables = OrderedDict()
                for output_feature_name, output_feature in output_features.items():
                    tables[output_feature_name] = [
                        [output_feature_name] + metrics_names[output_feature_name]
                    ]
                tables[COMBINED] = [[COMBINED, LOSS]]

                # eval metrics on train
                self.evaluation(
                    model,
                    training_set,
                    'train',
                    progress_tracker.train_metrics,
                    tables,
                    self.eval_batch_size,
                )

                self.write_epoch_summary(
                    summary_writer=train_summary_writer,
                    metrics=progress_tracker.train_metrics,
                    step=progress_tracker.epoch,
                )

                if validation_set is not None and len(validation_set) > 0:
                    self.callback(lambda c: c.on_validation_start(self, progress_tracker, save_path))

                    # eval metrics on validation set
                    self.evaluation(
                        model,
                        validation_set,
                        'vali',
                        progress_tracker.vali_metrics,
                        tables,
                        self.eval_batch_size,
                    )

                    self.write_epoch_summary(
                        summary_writer=validation_summary_writer,
                        metrics=progress_tracker.vali_metrics,
                        step=progress_tracker.epoch,
                    )

                    self.callback(lambda c: c.on_validation_end(self, progress_tracker, save_path))

                if test_set is not None and len(test_set) > 0:
                    self.callback(lambda c: c.on_test_start(self, progress_tracker, save_path))

                    # eval metrics on test set
                    self.evaluation(
                        model,
                        test_set,
                        TEST,
                        progress_tracker.test_metrics,
                        tables,
                        self.eval_batch_size,
                    )

                    self.write_epoch_summary(
                        summary_writer=test_summary_writer,
                        metrics=progress_tracker.test_metrics,
                        step=progress_tracker.epoch,
                    )

                    self.callback(lambda c: c.on_test_end(self, progress_tracker, save_path))

                elapsed_time = (time.time() - start_time) * 1000.0

                if self.is_coordinator():
                    logger.info('Took {time}'.format(
                        time=time_utils.strdelta(elapsed_time)))

                # metric prints
                if self.is_coordinator():
                    for output_feature, table in tables.items():
                        logger.info(
                            tabulate(
                                table,
                                headers='firstrow',
                                tablefmt='fancy_grid',
                                floatfmt='.4f'
                            )
                        )

                # ================ Validation Logic ================
                if should_validate:
                    should_break = self.check_progress_on_validation(
                        model,
                        progress_tracker,
                        self.validation_field,
                        self.validation_metric,
                        model_weights_path,
                        model_hyperparameters_path,
                        self.reduce_learning_rate_on_plateau,
                        self.reduce_learning_rate_on_plateau_patience,
                        self.reduce_learning_rate_on_plateau_rate,
                        self.reduce_learning_rate_eval_metric,
                        self.reduce_learning_rate_eval_split,
                        self.increase_batch_size_on_plateau,
                        self.increase_batch_size_on_plateau_patience,
                        self.increase_batch_size_on_plateau_rate,
                        self.increase_batch_size_on_plateau_max,
                        self.increase_batch_size_eval_metric,
                        self.increase_batch_size_eval_split,
                        self.early_stop,
                        self.skip_save_model,
                    )
                    if should_break:
                        break
                else:
                    # there's no validation, so we save the model at each iteration
                    if self.is_coordinator() and not self.skip_save_model:
                        model.save_weights(model_weights_path)

                # ========== Save training progress ==========
                if self.is_coordinator():
                    if not self.skip_save_progress:
                        checkpoint_manager.save()
                        progress_tracker.save(
                            os.path.join(
                                save_path,
                                TRAINING_PROGRESS_TRACKER_FILE_NAME
                            )
                        )
                    logger.info('')

                self.callback(lambda c: c.on_epoch_end(self, progress_tracker, save_path))

        if train_summary_writer is not None:
            train_summary_writer.close()
        if validation_summary_writer is not None:
            validation_summary_writer.close()
        if test_summary_writer is not None:
            test_summary_writer.close()

        return (
            model,
            progress_tracker.train_metrics,
            progress_tracker.vali_metrics,
            progress_tracker.test_metrics
        )

    def train_online(
            self,
            model,
            dataset,
    ):
        with dataset.initialize_batcher(
            batch_size=self.batch_size,
            should_shuffle=self.should_shuffle,
            shuffle_buffer_size=self.shuffle_buffer_size,
            horovod=self.horovod
        ) as batcher:

            # training step loop
            progress_bar = tqdm(
                desc='Trainining online',
                total=batcher.steps_per_epoch,
                file=sys.stdout,
                disable=is_progressbar_disabled()
            )

            while not batcher.last_batch():
                batch = batcher.next_batch()
                inputs = {
                    i_feat.feature_name: batch[i_feat.proc_column]
                    for i_feat in model.input_features.values()
                }
                targets = {
                    o_feat.feature_name: batch[o_feat.proc_column]
                    for o_feat in model.output_features.values()
                }

                model.train_step(
                    self.optimizer,
                    inputs,
                    targets,
                    self.regularization_lambda
                )

                progress_bar.update(1)

            progress_bar.close()
        return model

    @property
    def validation_field(self):
        return self._validation_field

    @property
    def validation_metric(self):
        return self._validation_metric

    def append_metrics(self, model, dataset_name, results, metrics_log,
                       tables):
        for output_feature in model.output_features:
            scores = [dataset_name]

            # collect metric names based on output features metrics to
            # ensure consistent order of reporting metrics
            metric_names = model.output_features[output_feature] \
                .metric_functions.keys()

            for metric in metric_names:
                score = results[output_feature][metric]
                metrics_log[output_feature][metric].append(score)
                scores.append(score)

            tables[output_feature].append(scores)

        metrics_log[COMBINED][LOSS].append(results[COMBINED][LOSS])
        tables[COMBINED].append([dataset_name, results[COMBINED][LOSS]])

        return metrics_log, tables

    def evaluation(
            self,
            model,
            dataset,
            dataset_name,
            metrics_log,
            tables,
            batch_size=128,
    ):
        predictor = Predictor(
            batch_size=batch_size, horovod=self.horovod, debug=self.debug
        )
        metrics, predictions = predictor.batch_evaluation(
            model,
            dataset,
            collect_predictions=False,
            dataset_name=dataset_name
        )

        self.append_metrics(model, dataset_name, metrics, metrics_log, tables)

        return metrics_log, tables

    def check_progress_on_validation(
            self,
            model,
            progress_tracker,
            validation_output_feature_name,
            validation_metric,
            model_weights_path,
            model_hyperparameters_path,
            reduce_learning_rate_on_plateau,
            reduce_learning_rate_on_plateau_patience,
            reduce_learning_rate_on_plateau_rate,
            reduce_learning_rate_eval_metric,
            reduce_learning_rate_eval_split,
            increase_batch_size_on_plateau,
            increase_batch_size_on_plateau_patience,
            increase_batch_size_on_plateau_rate,
            increase_batch_size_on_plateau_max,
            increase_batch_size_eval_metric,
            increase_batch_size_eval_split,
            early_stop,
            skip_save_model
    ):
        should_break = False
        # record how long its been since an improvement
        improved = get_improved_fun(validation_metric)
        if improved(
                progress_tracker.vali_metrics[validation_output_feature_name][
                    validation_metric][-1],
                progress_tracker.best_eval_metric
        ):
            progress_tracker.last_improvement_epoch = progress_tracker.epoch
            progress_tracker.best_eval_metric = progress_tracker.vali_metrics[
                validation_output_feature_name][validation_metric][-1]
            if self.is_coordinator() and not skip_save_model:
                model.save_weights(model_weights_path)
                logger.info(
                    'Validation {} on {} improved, model saved'.format(
                        validation_metric,
                        validation_output_feature_name
                    )
                )

        progress_tracker.last_improvement = (
                progress_tracker.epoch - progress_tracker.last_improvement_epoch
        )
        if progress_tracker.last_improvement != 0 and self.is_coordinator():
            logger.info(
                'Last improvement of {} validation {} '
                'happened {} epoch{} ago'.format(
                    validation_output_feature_name,
                    validation_metric,
                    progress_tracker.last_improvement,
                    '' if progress_tracker.last_improvement == 1 else 's'
                )
            )

        # ========== Reduce Learning Rate Plateau logic ========
        if reduce_learning_rate_on_plateau > 0:
            self.reduce_learning_rate(
                progress_tracker,
                validation_output_feature_name,
                reduce_learning_rate_on_plateau,
                reduce_learning_rate_on_plateau_patience,
                reduce_learning_rate_on_plateau_rate,
                reduce_learning_rate_eval_metric,
                reduce_learning_rate_eval_split
            )
            progress_tracker.last_learning_rate_reduction = (
                    progress_tracker.epoch -
                    progress_tracker.last_learning_rate_reduction_epoch
            )
            if (
                    progress_tracker.last_learning_rate_reduction > 0
                    and
                    progress_tracker.last_reduce_learning_rate_eval_metric_improvement > 0
                    and
                    not progress_tracker.num_reductions_learning_rate >= reduce_learning_rate_on_plateau
            ):
                logger.info(
                    'Last learning rate reduction '
                    'happened {} epoch{} ago, '
                    'improvement of {} {} {} '
                    'happened {} epoch{} ago'
                    ''.format(
                        progress_tracker.last_learning_rate_reduction,
                        '' if progress_tracker.last_learning_rate_reduction == 1 else 's',
                        validation_output_feature_name,
                        reduce_learning_rate_eval_split,
                        reduce_learning_rate_eval_metric,
                        progress_tracker.last_reduce_learning_rate_eval_metric_improvement,
                        '' if progress_tracker.last_reduce_learning_rate_eval_metric_improvement == 1 else 's',
                    )
                )

        # ========== Increase Batch Size Plateau logic =========
        if increase_batch_size_on_plateau > 0:
            self.increase_batch_size(
                progress_tracker,
                validation_output_feature_name,
                increase_batch_size_on_plateau,
                increase_batch_size_on_plateau_patience,
                increase_batch_size_on_plateau_rate,
                increase_batch_size_on_plateau_max,
                increase_batch_size_eval_metric,
                increase_batch_size_eval_split
            )
            progress_tracker.last_increase_batch_size = (
                    progress_tracker.epoch -
                    progress_tracker.last_increase_batch_size_epoch
            )
            if (
                    progress_tracker.last_increase_batch_size > 0
                    and
                    progress_tracker.last_increase_batch_size_eval_metric_improvement > 0
                    and
                    not progress_tracker.num_increases_batch_size >= increase_batch_size_on_plateau
                    and
                    not progress_tracker.batch_size >= increase_batch_size_on_plateau_max
            ):
                logger.info(
                    'Last batch size increase '
                    'happened {} epoch{} ago, '
                    'improvement of {} {} {} '
                    'happened {} epoch{} ago'.format(
                        progress_tracker.last_increase_batch_size,
                        '' if progress_tracker.last_increase_batch_size == 1 else 's',
                        validation_output_feature_name,
                        increase_batch_size_eval_split,
                        increase_batch_size_eval_metric,
                        progress_tracker.last_increase_batch_size_eval_metric_improvement,
                        '' if progress_tracker.last_increase_batch_size_eval_metric_improvement == 1 else 's',
                    )
                )

        # ========== Early Stop logic ==========
        if 0 < early_stop <= progress_tracker.last_improvement:
            if self.is_coordinator():
                logger.info(
                    "\nEARLY STOPPING due to lack of "
                    "validation improvement, "
                    "it has been {0} epochs since last "
                    "validation improvement\n".format(
                        progress_tracker.epoch -
                        progress_tracker.last_improvement_epoch
                    )
                )
            should_break = True
        return should_break

    def set_epochs_to_1_or_quit(self, signum, frame):
        if not self.received_sigint:
            self.epochs = 1
            self.received_sigint = True
            logger.critical(
                '\nReceived SIGINT, will finish this epoch and then conclude '
                'the training'
            )
            logger.critical(
                'Send another SIGINT to immediately interrupt the process'
            )
        else:
            logger.critical('\nReceived a second SIGINT, will now quit')
            sys.exit(1)

    def quit_training(self, signum, frame):
        logger.critical('Received SIGQUIT, will kill training')
        sys.exit(1)

    def resume_training_progress_tracker(self, training_progress_tracker_path):
        if self.is_coordinator():
            logger.info('Resuming training of model: {0}'.format(
                training_progress_tracker_path
            ))
        progress_tracker = ProgressTracker.load(training_progress_tracker_path)
        return progress_tracker

    def initialize_training_metrics(self, output_features):
        train_metrics = OrderedDict()
        vali_metrics = OrderedDict()
        test_metrics = OrderedDict()

        for output_feature_name, output_feature in output_features.items():
            train_metrics[output_feature_name] = OrderedDict()
            vali_metrics[output_feature_name] = OrderedDict()
            test_metrics[output_feature_name] = OrderedDict()
            for metric in output_feature.metric_functions:
                train_metrics[output_feature_name][metric] = []
                vali_metrics[output_feature_name][metric] = []
                test_metrics[output_feature_name][metric] = []

        for metrics in [train_metrics, vali_metrics, test_metrics]:
            metrics[COMBINED] = {LOSS: []}

        return train_metrics, vali_metrics, test_metrics

    def get_metrics_names(self, output_features):
        metrics_names = {}
        for output_feature_name, output_feature in output_features.items():
            for metric in output_feature.metric_functions:
                metrics = metrics_names.get(output_feature_name, [])
                metrics.append(metric)
                metrics_names[output_feature_name] = metrics
        metrics_names[COMBINED] = [LOSS]
        return metrics_names

    def resume_weights_and_optimzier(
            self,
            model_weights_progress_path,
            checkpoint
    ):
        checkpoint.restore(
            tf.train.latest_checkpoint(model_weights_progress_path)
        )

    def reduce_learning_rate(
            self,
            progress_tracker,
            validation_output_feature_name,
            reduce_learning_rate_on_plateau,
            reduce_learning_rate_on_plateau_patience,
            reduce_learning_rate_on_plateau_rate,
            reduce_learning_rate_eval_metric=LOSS,
            reduce_learning_rate_eval_split=TRAINING
    ):
        if not (progress_tracker.num_reductions_learning_rate >=
                reduce_learning_rate_on_plateau):

            if reduce_learning_rate_eval_split == TRAINING:
                split_metrics = progress_tracker.train_metrics
            elif reduce_learning_rate_eval_split == VALIDATION:
                split_metrics = progress_tracker.vali_metrics
            else:  # if reduce_learning_rate_eval_split == TEST:
                split_metrics = progress_tracker.test_metrics

            validation_metric = reduce_learning_rate_eval_metric
            last_metric_value = split_metrics[validation_output_feature_name][
                validation_metric][-1]

            improved = get_improved_fun(validation_metric)
            is_improved = improved(
                last_metric_value,
                progress_tracker.best_reduce_learning_rate_eval_metric
            )
            if is_improved:
                # we update the best metric value and set it to the current one
                # and reset last improvement epoch count
                progress_tracker.best_reduce_learning_rate_eval_metric = last_metric_value
                progress_tracker.last_reduce_learning_rate_eval_metric_improvement = 0
            else:
                progress_tracker.last_reduce_learning_rate_eval_metric_improvement += 1
                if not is_improved and (
                        # learning rate reduction happened more than N epochs ago
                        progress_tracker.last_learning_rate_reduction >=
                        reduce_learning_rate_on_plateau_patience
                        and
                        # we had no improvement of the evaluation metric since more than N epochs ago
                        progress_tracker.last_reduce_learning_rate_eval_metric_improvement >=
                        reduce_learning_rate_on_plateau_patience
                ):
                    progress_tracker.learning_rate *= (
                        reduce_learning_rate_on_plateau_rate
                    )

                    if self.is_coordinator():
                        logger.info(
                            'PLATEAU REACHED, reducing learning rate to {} '
                            'due to lack of improvement of {} {} {}'.format(
                                progress_tracker.learning_rate,
                                validation_output_feature_name,
                                reduce_learning_rate_eval_split,
                                validation_metric,
                            )
                        )

                    progress_tracker.last_learning_rate_reduction_epoch = progress_tracker.epoch
                    progress_tracker.last_learning_rate_reduction = 0
                    progress_tracker.num_reductions_learning_rate += 1

                    if (progress_tracker.num_reductions_learning_rate >=
                            reduce_learning_rate_on_plateau):
                        if self.is_coordinator():
                            logger.info(
                                'Learning rate was already reduced '
                                '{} times, not reducing it anymore'.format(
                                    progress_tracker.num_reductions_learning_rate
                                )
                            )

    def increase_batch_size(
            self,
            progress_tracker,
            validation_output_feature_name,
            increase_batch_size_on_plateau,
            increase_batch_size_on_plateau_patience,
            increase_batch_size_on_plateau_rate,
            increase_batch_size_on_plateau_max,
            increase_batch_size_eval_metric=LOSS,
            increase_batch_size_eval_split=TRAINING
    ):
        if (not progress_tracker.num_increases_batch_size >=
                increase_batch_size_on_plateau
                and not progress_tracker.batch_size ==
                        increase_batch_size_on_plateau_max):

            if increase_batch_size_eval_split == TRAINING:
                split_metrics = progress_tracker.train_metrics
            elif increase_batch_size_eval_split == VALIDATION:
                split_metrics = progress_tracker.vali_metrics
            else:  # if increase_batch_size_eval_split == TEST:
                split_metrics = progress_tracker.test_metrics

            validation_metric = increase_batch_size_eval_metric
            last_metric_value = split_metrics[validation_output_feature_name][
                validation_metric][-1]

            improved = get_improved_fun(validation_metric)
            is_improved = improved(
                last_metric_value,
                progress_tracker.best_increase_batch_size_eval_metric
            )
            if is_improved:
                # We update the best metric value and set it to the current one, and reset last improvement epoch count
                progress_tracker.best_increase_batch_size_eval_metric = last_metric_value
                progress_tracker.last_increase_batch_size_eval_metric_improvement = 0
            else:
                progress_tracker.last_increase_batch_size_eval_metric_improvement += 1
                if not is_improved and (
                        # Batch size increase happened more than N epochs ago
                        progress_tracker.last_increase_batch_size >=
                        increase_batch_size_on_plateau_patience
                        and
                        # We had no improvement of the evaluation metric since more than N epochs ago
                        progress_tracker.last_increase_batch_size_eval_metric_improvement >=
                        increase_batch_size_on_plateau_patience
                ):
                    progress_tracker.batch_size = min(
                        (increase_batch_size_on_plateau_rate *
                         progress_tracker.batch_size),
                        increase_batch_size_on_plateau_max
                    )

                    if self.is_coordinator():
                        logger.info(
                            'PLATEAU REACHED, increasing batch size to {} '
                            'due to lack of improvement of {} {} {}'.format(
                                progress_tracker.batch_size,
                                validation_output_feature_name,
                                increase_batch_size_eval_split,
                                validation_metric,
                            )
                        )

                    progress_tracker.last_increase_batch_size_epoch = progress_tracker.epoch
                    progress_tracker.last_increase_batch_size = 0
                    progress_tracker.num_increases_batch_size += 1

                    if (progress_tracker.num_increases_batch_size >=
                            increase_batch_size_on_plateau):
                        if self.is_coordinator():
                            logger.info(
                                'Batch size was already increased '
                                '{} times, not increasing it anymore'.format(
                                    progress_tracker.num_increases_batch_size
                                )
                            )
                    elif (progress_tracker.batch_size >=
                          increase_batch_size_on_plateau_max):
                        if self.is_coordinator():
                            logger.info(
                                'Batch size was already increased '
                                '{} times, currently it is {}, '
                                'the maximum allowed'.format(
                                    progress_tracker.num_increases_batch_size,
                                    progress_tracker.batch_size
                                )
                            )

    def is_coordinator(self):
        if not self.horovod:
            return True
        return self.horovod.rank() == 0

    def callback(self, fn):
        if self.is_coordinator():
            for callback in self.callbacks:
                fn(callback)


class RemoteTrainer(Trainer):
    def __init__(
            self,
            gpus=None,
            gpu_memory_limit=None,
            allow_parallel_threads=True,
            **kwargs
    ):
        horovod = initialize_horovod()
        initialize_tensorflow(gpus=gpus,
                              gpu_memory_limit=gpu_memory_limit,
                              allow_parallel_threads=allow_parallel_threads,
                              horovod=horovod)
        super().__init__(horovod=horovod, **kwargs)

        # Only return results from rank 0 to reduce network overhead
        self.train = return_first(self.train)
        self.train_online = return_first(self.train_online)


class ProgressTracker:
    def __init__(
            self,
            epoch,
            batch_size,
            steps,
            last_improvement_epoch,
            last_learning_rate_reduction_epoch,
            last_increase_batch_size_epoch,
            best_eval_metric,
            best_reduce_learning_rate_eval_metric,
            last_reduce_learning_rate_eval_metric_improvement,
            best_increase_batch_size_eval_metric,
            last_increase_batch_size_eval_metric_improvement,
            learning_rate,
            num_reductions_learning_rate,
            num_increases_batch_size,
            train_metrics,
            vali_metrics,
            test_metrics,
            last_improvement,
            last_learning_rate_reduction,
            last_increase_batch_size
    ):
        self.batch_size = batch_size
        self.epoch = epoch
        self.steps = steps
        self.last_improvement_epoch = last_improvement_epoch
        self.last_improvement = last_improvement
        self.last_learning_rate_reduction_epoch = last_learning_rate_reduction_epoch
        self.last_learning_rate_reduction = last_learning_rate_reduction
        self.last_increase_batch_size_epoch = last_increase_batch_size_epoch
        self.last_increase_batch_size = last_increase_batch_size
        self.learning_rate = learning_rate
        self.best_eval_metric = best_eval_metric
        self.best_reduce_learning_rate_eval_metric = best_reduce_learning_rate_eval_metric
        self.last_reduce_learning_rate_eval_metric_improvement = last_reduce_learning_rate_eval_metric_improvement
        self.best_increase_batch_size_eval_metric = best_increase_batch_size_eval_metric
        self.last_increase_batch_size_eval_metric_improvement = last_increase_batch_size_eval_metric_improvement
        self.num_reductions_learning_rate = num_reductions_learning_rate
        self.num_increases_batch_size = num_increases_batch_size
        self.train_metrics = train_metrics
        self.vali_metrics = vali_metrics
        self.test_metrics = test_metrics

    def save(self, filepath):
        save_json(filepath, self.__dict__)

    @staticmethod
    def load(filepath):
        loaded = load_json(filepath)
        return ProgressTracker(**loaded)

    @property
    def log_metrics(self):
        log_metrics = {}
        for item_name in [
            "batch_size",
            "epoch",
            "steps",
            "last_improvement_epoch",
            "learning_rate",
            "best_valid_metric",
            "num_reductions_lr",
            "num_increases_bs",
            "train_metrics",
            "vali_metrics",
            "test_metrics"
        ]:
            try:
                item = getattr(self, item_name)
                if isinstance(item, dict):
                    for key in item:
                        if isinstance(item[key], dict):
                            for key2 in item[key]:
                                log_metrics[
                                    item_name + "." + key + "." + key2
                                ] = item[key][key2][-1]
                        else:
                            log_metrics[item_name + "." + key] = item[key][-1]
                elif item is not None:
                    log_metrics[item_name] = item
            except Exception:
                logger.info(f"skip logging '{item_name}'")
        return log_metrics



import copy
import logging
from collections import OrderedDict

import tensorflow as tf

from ludwig.combiners.combiners import get_combiner_class
from ludwig.constants import *
from ludwig.features.feature_registries import input_type_registry, \
    output_type_registry
from ludwig.utils.algorithms_utils import topological_sort_feature_dependencies
from ludwig.utils.data_utils import clear_data_cache
from ludwig.utils.misc_utils import get_from_registry

logger = logging.getLogger(__name__)


class ECD(tf.keras.Model):

    def __init__(
            self,
            input_features_def,
            combiner_def,
            output_features_def,
            random_seed=None,
    ):
        # Deep copy to prevent TensorFlow from hijacking the dicts within the config and
        # transforming them into _DictWrapper classes, which are not JSON serializable.
        self._input_features_df = copy.deepcopy(input_features_def)
        self._combiner_def = copy.deepcopy(combiner_def)
        self._output_features_df = copy.deepcopy(output_features_def)

        self._random_seed = random_seed

        if random_seed is not None:
            tf.random.set_seed(random_seed)

        super().__init__()

        # ================ Inputs ================
        self.input_features = build_inputs(
            input_features_def
        )

        # ================ Combiner ================
        logger.debug('Combiner {}'.format(combiner_def[TYPE]))
        combiner_class = get_combiner_class(combiner_def[TYPE])
        self.combiner = combiner_class(
            input_features=self.input_features,
            **combiner_def,
        )

        # ================ Outputs ================
        self.output_features = build_outputs(
            output_features_def,
            self.combiner
        )

        # ================ Combined loss metric ================
        self.eval_loss_metric = tf.keras.metrics.Mean()

        # After constructing all layers, clear the cache to free up memory
        clear_data_cache()


    def get_model_inputs(self, training=True):
        inputs = {
            input_feature_name: input_feature.create_input()
            for input_feature_name, input_feature in
            self.input_features.items()
        }

        if not training:
            return inputs

        targets = {
            output_feature_name: output_feature.create_input()
            for output_feature_name, output_feature in
            self.output_features.items()
        }
        return inputs, targets

    def get_connected_model(self, training=True, inputs=None):
        inputs = inputs or self.get_model_inputs(training)
        outputs = self.call(inputs)
        return tf.keras.Model(inputs=inputs, outputs=outputs)

    def save_savedmodel(self, save_path):
        keras_model = self.get_connected_model(training=False)
        keras_model.save(save_path)

    def call(self, inputs, training=None, mask=None):
        # parameter inputs is a dict feature_name -> tensor / ndarray
        # or
        # parameter (inputs, targets) where
        #   inputs is a dict feature_name -> tensor/ndarray
        #   targets is dict feature_name -> tensor/ndarray

        if isinstance(inputs, tuple):
            inputs, targets = inputs
        else:
            targets = None
        assert inputs.keys() == self.input_features.keys()

        encoder_outputs = {}
        for input_feature_name, input_values in inputs.items():
            encoder = self.input_features[input_feature_name]
            encoder_output = encoder(input_values, training=training,
                                     mask=mask)
            encoder_outputs[input_feature_name] = encoder_output

        combiner_outputs = self.combiner(encoder_outputs)

        output_logits = {}
        output_last_hidden = {}
        for output_feature_name, decoder in self.output_features.items():
            # use presence or absence of targets
            # to signal training or prediction
            decoder_inputs = (combiner_outputs, copy.copy(output_last_hidden))
            if targets is not None:
                # targets are only used during training,
                # during prediction they are omitted
                decoder_inputs = (decoder_inputs, targets[output_feature_name])

            decoder_outputs = decoder(
                decoder_inputs,
                training=training,
                mask=mask
            )
            output_logits[output_feature_name] = decoder_outputs
            output_last_hidden[output_feature_name] = decoder_outputs[
                'last_hidden']

        return output_logits

    def predictions(self, inputs, output_features=None):
        # check validity of output_features
        if output_features is None:
            of_list = self.output_features
        elif isinstance(output_features, str):
            if output_features == 'all':
                of_list = set(self.output_features.keys())
            elif output_features in self.output_features:
                of_list = [output_features]
            else:
                raise ValueError(
                    "'output_features' {} is not a valid for this model. "
                    "Available ones are: {}".format(
                        output_features, set(self.output_features.keys())
                    )
                )
        elif isinstance(output_features, list or set):
            if output_features.issubset(self.output_features):
                of_list = output_features
            else:
                raise ValueError(
                    "'output_features' {} must be a subset of "
                    "available features {}".format(
                        output_features, set(self.output_features.keys())
                    )
                )
        else:
            raise ValueError(
                "'output_features' must be None or a string or a list "
                "of output features"
            )

        outputs = self.call(inputs, training=False)

        predictions = {}
        for of_name in of_list:
            predictions[of_name] = self.output_features[of_name].predictions(
                outputs[of_name],
                training=False
            )

        return predictions

    @tf.function
    def train_step(self, optimizer, inputs, targets,
                   regularization_lambda=0.0):
        with tf.GradientTape() as tape:
            model_outputs = self((inputs, targets), training=True)
            loss, all_losses = self.train_loss(
                targets, model_outputs, regularization_lambda
            )
        optimizer.minimize_with_tape(
            tape, loss, self.trainable_variables
        )
        # grads = tape.gradient(loss, model.trainable_weights)
        # optimizer.apply_gradients(zip(grads, model.trainable_weights))
        return loss, all_losses

    @tf.function
    def evaluation_step(self, inputs, targets):
        predictions = self.predictions(inputs, output_features=None)
        self.update_metrics(targets, predictions)
        return predictions

    @tf.function
    def predict_step(self, inputs):
        return self.predictions(inputs, output_features=None)

    def train_loss(self, targets, predictions, regularization_lambda=0.0):
        train_loss = 0
        of_train_losses = {}
        for of_name, of_obj in self.output_features.items():
            of_train_loss = of_obj.train_loss(targets[of_name],
                                              predictions[of_name])
            train_loss += of_obj.loss['weight'] * of_train_loss
            of_train_losses[of_name] = of_train_loss

        for loss in self.losses:
            if hasattr(loss, "loss_name"):
                # this assumes that all losses with a loss_name are not
                # regularization losses and should be added
                # to the total loss as they are
                train_loss += loss
            else:
                # this assumes all unnamed losses are regularization losses
                train_loss += regularization_lambda * loss

        return train_loss, of_train_losses

    def eval_loss(self, targets, predictions):
        eval_loss = 0
        of_eval_losses = {}
        for of_name, of_obj in self.output_features.items():
            of_eval_loss = of_obj.eval_loss(
                targets[of_name], predictions[of_name]
            )
            eval_loss += of_obj.loss['weight'] * of_eval_loss
            of_eval_losses[of_name] = of_eval_loss
        eval_loss += sum(self.losses)  # regularization / other losses
        return eval_loss, of_eval_losses

    def update_metrics(self, targets, predictions):
        for of_name, of_obj in self.output_features.items():
            of_obj.update_metrics(targets[of_name], predictions[of_name])
        self.eval_loss_metric.update_state(
            self.eval_loss(targets, predictions)[0]
        )

    def get_metrics(self):
        all_of_metrics = {}
        for of_name, of_obj in self.output_features.items():
            all_of_metrics[of_name] = of_obj.get_metrics()
        all_of_metrics[COMBINED] = {
            LOSS: self.eval_loss_metric.result().numpy()
        }
        return all_of_metrics

    def reset_metrics(self):
        for of_obj in self.output_features.values():
            of_obj.reset_metrics()
        self.eval_loss_metric.reset_states()

    def collect_weights(
            self,
            tensor_names=None,
            **kwargs
    ):
        def recurse_weights(model, prefix=None):
            results = []
            for layer in model.layers:
                layer_prefix = f'{prefix}/{layer.name}' if prefix else layer.name
                if isinstance(layer, tf.keras.Model):
                    results += recurse_weights(layer, layer_prefix)
                else:
                    results += [(f'{layer_prefix}/{w.name}', w) for w in
                                layer.weights]
            return results

        connected_model = self.get_connected_model()
        weights = recurse_weights(connected_model)
        if tensor_names:
            # Check for bad tensor names
            weight_set = set(name for name, w in weights)
            for name in tensor_names:
                if name not in weight_set:
                    raise ValueError(
                        f'Tensor {name} not present in the model graph')

            # Filter the weights
            tensor_set = set(tensor_names)
            weights = [(name, w) for name, w in weights if name in tensor_set]

        return weights

    def get_args(self):
        return self._input_features_df, self._combiner_def, self._output_features_df, self._random_seed

    def __setstate__(self, newstate):
        self.set_weights(newstate['weights'])

    def __reduce__(self):
        args = self.get_args()
        state = {'weights': self.get_weights()}
        return ECD, args, state


def build_inputs(
        input_features_def,
        **kwargs
):
    input_features = OrderedDict()
    input_features_def = topological_sort_feature_dependencies(
        input_features_def)
    for input_feature_def in input_features_def:
        input_features[input_feature_def[NAME]] = build_single_input(
            input_feature_def,
            input_features,
            **kwargs
        )
    return input_features


def build_single_input(input_feature_def, other_input_features, **kwargs):
    logger.debug('Input {} feature {}'.format(
        input_feature_def[TYPE],
        input_feature_def[NAME]
    ))

    encoder_obj = None
    if input_feature_def.get(TIED, None) is not None:
        tied_input_feature_name = input_feature_def[TIED]
        if tied_input_feature_name in other_input_features:
            encoder_obj = other_input_features[
                tied_input_feature_name].encoder_obj

    input_feature_class = get_from_registry(
        input_feature_def[TYPE],
        input_type_registry
    )
    input_feature_obj = input_feature_class(input_feature_def, encoder_obj)

    return input_feature_obj


dynamic_length_encoders = {
    'rnn',
    'embed'
}


def build_outputs(
        output_features_def,
        combiner,
        **kwargs
):
    output_features_def = topological_sort_feature_dependencies(
        output_features_def)
    output_features = {}

    for output_feature_def in output_features_def:
        output_feature = build_single_output(
            output_feature_def,
            combiner,
            output_features,
            **kwargs
        )
        output_features[output_feature_def[NAME]] = output_feature

    return output_features


def build_single_output(
        output_feature_def,
        feature_hidden,
        other_output_features,
        **kwargs
):
    logger.debug('Output {} feature {}'.format(
        output_feature_def[TYPE],
        output_feature_def[NAME]
    ))

    output_feature_class = get_from_registry(
        output_feature_def[TYPE],
        output_type_registry
    )
    output_feature_obj = output_feature_class(output_feature_def)
    # weighted_train_mean_loss, weighted_eval_loss, output_tensors = output_feature_obj.concat_dependencies_and_build_output(
    #    feature_hidden,
    #    other_output_features,
    #    **kwargs
    # )

    return output_feature_obj



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Dropout
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.layers import Embedding

from ludwig.constants import TYPE
from ludwig.modules.initializer_modules import get_initializer
from ludwig.utils.data_utils import load_pretrained_embeddings

logger = logging.getLogger(__name__)


def embedding_matrix(
        vocab,
        embedding_size,
        representation='dense',
        embeddings_trainable=True,
        pretrained_embeddings=None,
        force_embedding_size=False,
        embedding_initializer=None,
):
    vocab_size = len(vocab)
    if representation == 'dense':
        if pretrained_embeddings is not None and pretrained_embeddings is not False:
            embeddings_matrix = load_pretrained_embeddings(
                pretrained_embeddings, vocab
            )
            if embeddings_matrix.shape[-1] != embedding_size:
                raise ValueError(
                    'The size of the pretrained embeddings is {}, '
                    'but the specified embedding_size is {}. '
                    'Please change the embedding_size accordingly.'.format(
                        embeddings_matrix.shape[-1],
                        embedding_size
                    ))
            embedding_initializer_obj = tf.constant(embeddings_matrix,
                                                    dtype=tf.float32)

        else:
            if vocab_size < embedding_size and not force_embedding_size:
                logger.info(
                    '  embedding_size ({}) is greater than vocab_size ({}). '
                    'Setting embedding size to be equal to vocab_size.'.format(
                        embedding_size, vocab_size
                    ))
                embedding_size = vocab_size

            if embedding_initializer is not None:
                embedding_initializer_obj_ref = get_initializer(
                    embedding_initializer)
            else:
                embedding_initializer_obj_ref = get_initializer(
                    {TYPE: 'uniform', 'minval': -1.0, 'maxval': 1.0})
            embedding_initializer_obj = embedding_initializer_obj_ref(
                [vocab_size, embedding_size])

        embeddings = tf.Variable(
            embedding_initializer_obj,
            trainable=embeddings_trainable,
            name='embeddings'
        )

    elif representation == 'sparse':
        embedding_size = vocab_size
        embeddings = tf.Variable(
            get_initializer('identity')([vocab_size, embedding_size]),
            trainable=False,
            name='embeddings'
        )

    else:
        raise Exception(
            'Embedding representation {} not supported.'.format(
                representation))

    return embeddings, embedding_size


def embedding_matrix_on_device(
        vocab,
        embedding_size,
        representation='dense',
        embeddings_trainable=True,
        pretrained_embeddings=None,
        force_embedding_size=False,
        embeddings_on_cpu=False,
        embedding_initializer=None
):
    if embeddings_on_cpu:
        with tf.device('/cpu:0'):
            embeddings, embedding_size = embedding_matrix(
                vocab,
                embedding_size,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                force_embedding_size=force_embedding_size,
                embedding_initializer=embedding_initializer
            )
    else:
        embeddings, embedding_size = embedding_matrix(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embedding_initializer=embedding_initializer
        )

    # logger.debug('  embeddings: {0}'.format(embeddings))

    return embeddings, embedding_size


class Embed(Layer):
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=False,
            embeddings_on_cpu=False,
            dropout=0.0,
            embedding_initializer=None,
            embedding_regularizer=None
    ):
        super().__init__()
        self.supports_masking = True

        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )

        if embedding_regularizer:
            embedding_regularizer_obj = tf.keras.regularizers.get(
                embedding_regularizer)
            self.add_loss(lambda: embedding_regularizer_obj(self.embeddings))

        if dropout > 0:
            self.dropout = Dropout(dropout)
        else:
            self.dropout = None

    def call(self, inputs, training=None, mask=None):
        embedded = tf.nn.embedding_lookup(
            self.embeddings, inputs, name='embeddings_lookup'
        )

        if self.dropout:
            embedded = self.dropout(embedded, training=training)

        return embedded


class EmbedWeighted(Layer):
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=False,
            embeddings_on_cpu=False,
            dropout=0.0,
            embedding_initializer=None,
            embedding_regularizer=None
    ):
        super().__init__()

        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )
        self.vocab_length = len(vocab)

        if embedding_regularizer:
            embedding_regularizer_obj = tf.keras.regularizers.get(
                embedding_regularizer)
            self.add_loss(lambda: embedding_regularizer_obj(self.embeddings))

        if dropout > 0:
            self.dropout = Dropout(dropout)
        else:
            self.dropout = None

    def call(self, inputs, training=None, mask=None):
        signed_input = tf.cast(tf.sign(tf.abs(inputs)), tf.int32)
        multiple_hot_indexes = tf.multiply(
            signed_input,
            tf.constant(np.array([range(self.vocab_length)], dtype=np.int32))
        )
        embedded = tf.nn.embedding_lookup(
            self.embeddings, multiple_hot_indexes, name='embeddings_lookup'
        )

        # Get the multipliers to embeddings
        weights_mask = tf.expand_dims(inputs, -1)
        weighted_embedded = tf.multiply(embedded, weights_mask)

        embedded_reduced = tf.reduce_sum(weighted_embedded, 1)

        if self.dropout:
            embedded_reduced = self.dropout(embedded_reduced,
                                            training=training)

        return embedded_reduced


class EmbedSparse(Layer):
    def __init__(
            self,
            vocab,
            embedding_size=50,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=False,
            embeddings_on_cpu=False,
            dropout=0.0,
            embedding_initializer=None,
            embedding_regularizer=None,
            reduce_output='sum'
    ):
        super().__init__()

        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )

        if embedding_regularizer:
            embedding_regularizer_obj = tf.keras.regularizers.get(
                embedding_regularizer)
            self.add_loss(lambda: embedding_regularizer_obj(self.embeddings))

        if dropout > 0:
            self.dropout = Dropout(dropout)
        else:
            self.dropout = None

        self.reduce_output = reduce_output

    def call(self, inputs, training=None, mask=None):
        idx = tf.where(tf.equal(inputs, True))

        sparse_multiple_hot_indexes = tf.SparseTensor(
            idx,
            idx[:, 1],
            tf.shape(inputs, out_type=tf.int64)
        )

        embedded_reduced = tf.nn.embedding_lookup_sparse(
            self.embeddings,
            sparse_multiple_hot_indexes,
            sp_weights=None,
            combiner=self.reduce_output
        )

        if self.dropout:
            embedded_reduced = self.dropout(
                embedded_reduced, training=training
            )

        return embedded_reduced


class EmbedSequence(Layer):
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=False,
            embeddings_on_cpu=False,
            dropout=0.0,
            embedding_initializer=None,
            embedding_regularizer=None
    ):
        super().__init__()
        self.supports_masking = True

        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )

        if embedding_regularizer:
            embedding_regularizer_obj = tf.keras.regularizers.get(
                embedding_regularizer)
            self.add_loss(lambda: embedding_regularizer_obj(self.embeddings))

        if dropout > 0:
            self.dropout = Dropout(dropout)
        else:
            self.dropout = None

    def call(self, inputs, training=None, mask=None):
        embedded = tf.nn.embedding_lookup(
            self.embeddings, inputs, name='embeddings_lookup'
        )

        if mask is not None:
            mask_matrix = tf.cast(
                tf.expand_dims(mask, -1),
                dtype=tf.float32
            )
            embedded = tf.multiply(embedded, mask_matrix)

        if self.dropout:
            embedded = self.dropout(embedded, training=training)

        return embedded


class TokenAndPositionEmbedding(Layer):
    def __init__(self,
                 max_length,
                 vocab,
                 embedding_size,
                 representation='dense',
                 embeddings_trainable=True,
                 pretrained_embeddings=None,
                 force_embedding_size=False,
                 embeddings_on_cpu=False,
                 dropout=0.0,
                 embedding_initializer=None,
                 embedding_regularizer=None
                 ):
        super().__init__()
        self.token_embed = EmbedSequence(
            vocab=vocab,
            embedding_size=embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=embedding_initializer,
            embedding_regularizer=embedding_regularizer
        )
        self.position_embed = Embedding(
            input_dim=max_length,
            output_dim=self.token_embed.embedding_size
        )

    def call(self, inputs, training=None, mask=None):
        max_length = tf.shape(inputs)[-1]
        positions = tf.range(start=0, limit=max_length, delta=1)
        positions_hidden = self.position_embed(positions)
        token_hidden = self.token_embed(inputs)
        return token_hidden + positions_hidden



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os.path

import pytest

from ludwig.hyperopt.execution import get_build_hyperopt_executor
from ludwig.hyperopt.results import HyperoptResults
from ludwig.hyperopt.run import hyperopt
from ludwig.hyperopt.sampling import (get_build_hyperopt_sampler)
from ludwig.hyperopt.run import update_hyperopt_params_with_defaults
from ludwig.utils.defaults import merge_with_defaults, ACCURACY
from ludwig.utils.tf_utils import get_available_gpus_cuda_string
from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import text_feature

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)

HYPEROPT_CONFIG = {
    "parameters": {
        "training.learning_rate": {
            "type": "float",
            "low": 0.0001,
            "high": 0.1,
            "space": "log",
            "steps": 3,
        },
        "combiner.num_fc_layers": {
            "type": "int",
            "low": 1,
            "high": 4,
            "space": "linear",
            "steps": 3,
        },
       "combiner.fc_layers" : {
            'type': 'category',
            'values': [
                [{'fc_size': 512}, {'fc_size': 256}],
                [{'fc_size': 512}],
                [{'fc_size': 256}]
            ]
        },
        
        "utterance.cell_type": {
            "type": "category",
            "values": ["rnn", "gru"]
        },
        "utterance.bidirectional": {
            "type": "category",
            "values": [True, False]
        }
    },
    "goal": "minimize"
}

SAMPLERS = [
    {"type": "grid"},
    {"type": "random", "num_samples": 5},
    {"type": "pysot", "num_samples": 5},
]

EXECUTORS = [
    {"type": "serial"},
    {"type": "parallel", "num_workers": 4},
    {"type": "fiber", "num_workers": 4},
]


@pytest.mark.distributed
@pytest.mark.parametrize('sampler', SAMPLERS)
@pytest.mark.parametrize('executor', EXECUTORS)
def test_hyperopt_executor(sampler, executor, csv_filename,
                           validate_output_feature=False,
                           validation_metric=None):
    if executor['type'] == 'fiber' and sampler['type'] == 'grid':
        # This test is very slow and doesn't give us additional converage
        pytest.skip('Skipping Fiber grid search')

    input_features = [
        text_feature(name="utterance", cell_type="lstm", reduce_output="sum"),
        category_feature(vocab_size=2, reduce_input="sum")]

    output_features = [category_feature(vocab_size=2, reduce_input="sum")]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "num_fc_layers": 2},
        "training": {"epochs": 2, "learning_rate": 0.001}
    }

    config = merge_with_defaults(config)

    hyperopt_config = HYPEROPT_CONFIG.copy()

    if validate_output_feature:
        hyperopt_config['output_feature'] = output_features[0]['name']
    if validation_metric:
        hyperopt_config['validation_metric'] = validation_metric

    update_hyperopt_params_with_defaults(hyperopt_config)

    parameters = hyperopt_config["parameters"]
    split = hyperopt_config["split"]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config["metric"]
    goal = hyperopt_config["goal"]

    hyperopt_sampler = get_build_hyperopt_sampler(
        sampler["type"])(goal, parameters, **sampler)

    hyperopt_executor = get_build_hyperopt_executor(executor["type"])(
        hyperopt_sampler, output_feature, metric, split, **executor)

    hyperopt_executor.execute(config,
                              dataset=rel_path,
                              gpus=get_available_gpus_cuda_string())


@pytest.mark.distributed
def test_hyperopt_executor_with_metric(csv_filename):
    test_hyperopt_executor({"type": "random", "num_samples": 2},
                           {"type": "serial"},
                           csv_filename,
                           validate_output_feature=True,
                           validation_metric=ACCURACY)


@pytest.mark.distributed
@pytest.mark.parametrize('samplers', SAMPLERS)
def test_hyperopt_run_hyperopt(csv_filename, samplers):
    input_features = [
        text_feature(name="utterance", cell_type="lstm", reduce_output="sum"),
        category_feature(vocab_size=2, reduce_input="sum")]

    output_features = [category_feature(vocab_size=2, reduce_input="sum")]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "num_fc_layers": 2},
        "training": {"epochs": 2, "learning_rate": 0.001}
    }

    output_feature_name = output_features[0]['name']

    
    hyperopt_configs = {
        "parameters": {
            "training.learning_rate": {
                "type": "float",
                "low": 0.0001,
                "high": 0.01,
                "space": "log",
                "steps": 3,
            },
            output_feature_name + ".fc_layers": {
                'type': 'category',
                'values': [
                    [{'fc_size': 512}, {'fc_size': 256}],
                    [{'fc_size': 512}],
                    [{'fc_size': 256}]
                ]
            },
            output_feature_name + ".fc_size": {
                "type": "int",
                "low": 32,
                "high": 256,
                "steps": 5
            },
            output_feature_name + ".num_fc_layers": {
                'type': 'int',
                'low': 1,
                'high': 5,
                'space': 'linear',
                'steps': 4
            }
        },
        "goal": "minimize",
        'output_feature': output_feature_name,
        'validation_metrics': 'loss',
        'executor': {'type': 'serial'},
        'sampler': {'type': samplers["type"], 'num_samples': 2}
    }

    # add hyperopt parameter space to the config
    config['hyperopt'] = hyperopt_configs

    hyperopt_results = hyperopt(
        config,
        dataset=rel_path,
        output_directory='results_hyperopt'
    )

    # check for return results
    assert isinstance(hyperopt_results, HyperoptResults)

    # check for existence of the hyperopt statistics file
    assert os.path.isfile(
        os.path.join('results_hyperopt', 'hyperopt_statistics.json')
    )

    if os.path.isfile(
        os.path.join('results_hyperopt', 'hyperopt_statistics.json')
    ):
        os.remove( 
            os.path.join('results_hyperopt', 'hyperopt_statistics.json')
        )


@pytest.mark.distributed
def test_hyperopt_executor_get_metric_score():
    executor = EXECUTORS[0]
    output_feature = "of_name"
    split = 'test'

    train_stats = {
        'training': {
            output_feature: {
                'loss': [0.58760345, 1.5066891],
                'accuracy': [0.6666667, 0.33333334],
                'hits_at_k': [1.0, 1.0]
            },
            'combined': {
                'loss': [0.58760345, 1.5066891]
            }
        },
        'validation': {
            output_feature: {
                'loss': [0.30233705, 2.6505466],
                'accuracy': [1.0, 0.0],
                'hits_at_k': [1.0, 1.0]
            },
            'combined': {
                'loss': [0.30233705, 2.6505466]
            }
        },
        'test': {
            output_feature: {
                'loss': [1.0876318, 1.4353828],
                'accuracy': [0.7, 0.5],
                'hits_at_k': [1.0, 1.0]
            },
            'combined': {
                'loss': [1.0876318, 1.4353828]
            }
        }
    }

    eval_stats = {
        output_feature: {
            'loss': 1.4353828,
            'accuracy': 0.5,
            'hits_at_k': 1.0,
            'overall_stats': {
                'token_accuracy': 1.0,
                'avg_precision_macro': 1.0,
                'avg_recall_macro': 1.0,
                'avg_f1_score_macro': 1.0,
                'avg_precision_micro': 1.0,
                'avg_recall_micro': 1.0,
                'avg_f1_score_micro': 1.0,
                'avg_precision_weighted': 1.0,
                'avg_recall_weighted': 1.0,
                'avg_f1_score_weighted': 1.0,
                'kappa_score': 0.6
            },
            'combined': {'loss': 1.4353828}
        }
    }

    metric = 'loss'
    hyperopt_executor = get_build_hyperopt_executor(executor["type"])(
        None, output_feature, metric, split, **executor)
    score = hyperopt_executor.get_metric_score(train_stats, eval_stats)
    assert score == 1.0876318

    metric = 'accuracy'
    hyperopt_executor = get_build_hyperopt_executor(executor["type"])(
        None, output_feature, metric, split, **executor)
    score = hyperopt_executor.get_metric_score(train_stats, eval_stats)
    assert score == 0.7

    metric = 'overall_stats.kappa_score'
    hyperopt_executor = get_build_hyperopt_executor(executor["type"])(
        None, output_feature, metric, split, **executor)
    score = hyperopt_executor.get_metric_score(train_stats, eval_stats)
    assert score == 0.6



import logging
from collections import namedtuple

import pytest

from ludwig.api import LudwigModel
from ludwig.constants import *
from tests.integration_tests.utils import binary_feature
from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import numerical_feature
from tests.integration_tests.utils import text_feature

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)

TestCase = namedtuple('TestCase', 'output_features validation_metrics')


# output_features: output features to test
# validation_metrics: relevant metrics for the output feature


# test for single output feature
@pytest.mark.parametrize(
    'test_case',
    [
        TestCase(
            [numerical_feature()],
            ['loss', 'mean_squared_error', 'mean_absolute_error', 'r2']
        ),
        TestCase(
            [binary_feature()],
            ['loss', 'accuracy']
        ),
        TestCase(
            [category_feature()],
            ['loss', 'accuracy', 'hits_at_k']
        ),
        TestCase(
            [text_feature()],
            ['loss', 'token_accuracy', 'last_accuracy', 'edit_distance',
             'perplexity']
        )
    ]
)
def test_validation_metrics(test_case: TestCase, csv_filename: str):
    # setup test scenarios
    test_scenarios = []
    for output_feature in test_case.output_features:
        # single output feature capture feature specific metrics
        of_name = output_feature[COLUMN]
        for metric in test_case.validation_metrics:
            test_scenarios.append((of_name, metric))
            if len(test_case.output_features) == 1:
                # it shoudl work when there's only one output feature
                # and the metric applyys to the output feature type,
                # the output feature name should be replacing combined
                # and a warning should be printed about the substitution
                test_scenarios.append(('combined', metric))

    # add standard test for combined
    test_scenarios.append(('combined', 'loss'))

    # setup features for the test
    input_features = [numerical_feature(),
                      category_feature(),
                      binary_feature()]
    output_features = test_case.output_features

    # generate training data
    training_data = generate_data(
        input_features,
        output_features,
        filename=csv_filename
    )

    # loop through scenarios
    for validation_field, validation_metric in test_scenarios:
        # setup config
        config = {
            'input_features': input_features,
            'output_features': output_features,
            'training': {
                'epochs': 3,
                'validation_field': validation_field,
                'validation_metric': validation_metric
            }
        }

        model = LudwigModel(config)
        model.train(
            dataset=training_data,
            skip_save_training_description=True,
            skip_save_training_statistics=True,
            skip_save_log=True,
            skip_save_model=True,
            skip_save_processed_input=True,
            skip_save_progress=True
        )


# test for multiple output features
@pytest.mark.parametrize(
    'test_case',
    [
        TestCase(
            [numerical_feature(), numerical_feature()],
            []
        ),
        TestCase(
            [category_feature(), numerical_feature()],
            []
        )
    ]
)
def test_validation_metrics_mulitiple_output(test_case: TestCase,
                                             csv_filename: str):
    test_validation_metrics(test_case, csv_filename)


# negative test for invalid metric name
@pytest.mark.parametrize(
    'test_case',
    [
        TestCase(
            [numerical_feature()],
            ['invalid_metric']
        )
    ]
)
def test_validation_invalid_metric(test_case: TestCase,
                                   csv_filename: str):
    # this should generate ValueError Exception
    try:
        test_validation_metrics(test_case, csv_filename)
        raise RuntimeError(
            'test_validation_metrics() should have raised ValueError '
            'but did not.'
        )
    except ValueError:
        pass



import logging
import os
import re
import uuid
from pathlib import Path

from ludwig.constants import CHECKSUM, META, TRAINING, TEST, VALIDATION
from ludwig.data.cache.util import calculate_checksum
from ludwig.utils import data_utils
from ludwig.utils.fs_utils import path_exists, delete

logger = logging.getLogger(__name__)


def alphanum(v):
    """Filters a string to only its alphanumeric characters."""
    return re.sub(r'\W+', '', v)


class DatasetCache:
    def __init__(self, config, checksum, cache_map, dataset_manager):
        self.config = config
        self.checksum = checksum
        self.cache_map = cache_map
        self.dataset_manager = dataset_manager

    def get(self):
        training_set_metadata_fp = self.cache_map[META]
        if not path_exists(training_set_metadata_fp):
            return None

        cache_training_set_metadata = data_utils.load_json(
            training_set_metadata_fp
        )

        cached_training_set = self.cache_map[TRAINING] \
            if path_exists(self.cache_map[TRAINING]) else None

        cached_test_set = self.cache_map[TEST] \
            if path_exists(self.cache_map[TEST]) else None

        cached_validation_set = self.cache_map[VALIDATION] \
            if path_exists(self.cache_map[VALIDATION]) else None

        valid = self.checksum == cache_training_set_metadata.get(CHECKSUM) and \
            cached_training_set is not None

        return valid, \
            cache_training_set_metadata, \
            cached_training_set, \
            cached_test_set, \
            cached_validation_set

    def put(self, training_set, test_set, validation_set, training_set_metadata):
        logger.info('Writing preprocessed training set cache')
        training_set = self.dataset_manager.save(
            self.cache_map[TRAINING],
            training_set,
            self.config,
            training_set_metadata,
            TRAINING,
        )

        if test_set is not None:
            logger.info('Writing preprocessed test set cache')
            test_set = self.dataset_manager.save(
                self.cache_map[TEST],
                test_set,
                self.config,
                training_set_metadata,
                TEST,
            )

        if validation_set is not None:
            logger.info('Writing preprocessed validation set cache')
            validation_set = self.dataset_manager.save(
                self.cache_map[VALIDATION],
                validation_set,
                self.config,
                training_set_metadata,
                VALIDATION,
            )

        logger.info('Writing train set metadata')
        data_utils.save_json(
            self.cache_map[META],
            training_set_metadata
        )

        return training_set, test_set, validation_set, training_set_metadata

    def delete(self):
        for fname in self.cache_map.values():
            if path_exists(fname):
                delete(fname)


class CacheManager:
    def __init__(self, dataset_manager, cache_dir=None):
        self._dataset_manager = dataset_manager
        self._cache_dir = cache_dir

    def get_dataset_cache(self, config, dataset=None, training_set=None, test_set=None, validation_set=None):
        if dataset is not None:
            key = self.get_cache_key(dataset, config)
            cache_map = {
                META: self.get_cache_path(dataset, key, META, 'json'),
                TRAINING: self.get_cache_path(dataset, key, TRAINING),
                TEST: self.get_cache_path(dataset, key, TEST),
                VALIDATION: self.get_cache_path(dataset, key, VALIDATION),
            }
            return DatasetCache(config, key, cache_map, self._dataset_manager)
        else:
            key = self.get_cache_key(training_set, config)
            cache_map = {
                META: self.get_cache_path(training_set, key, META, 'json'),
                TRAINING: self.get_cache_path(training_set, key, TRAINING),
                TEST: self.get_cache_path(test_set, key, TEST),
                VALIDATION: self.get_cache_path(validation_set, key, VALIDATION),
            }
            return DatasetCache(config, key, cache_map, self._dataset_manager)

    def get_cache_key(self, dataset, config):
        if not isinstance(dataset, str):
            # TODO(travis): could try hashing the in-memory dataset, but this is tricky for Dask
            return str(uuid.uuid1())
        return calculate_checksum(dataset, config)

    def get_cache_path(self, dataset, key, tag, ext=None):
        if not isinstance(dataset, str):
            dataset = None

        if self._cache_dir is None and dataset is not None:
            # Use the input dataset filename (minus the extension) as the cache path
            stem = Path(dataset).stem
        else:
            # To avoid collisions across different directories, we use the unique checksum
            # as the cache path
            stem = alphanum(key)

        ext = ext or self.data_format
        cache_fname = f'{stem}.{tag}.{ext}'
        return os.path.join(self.get_cache_directory(dataset), cache_fname)

    def get_cache_directory(self, input_fname):
        if self._cache_dir is None:
            if input_fname is not None:
                return os.path.dirname(input_fname)
            return '.'
        return self._cache_dir

    def can_cache(self, skip_save_processed_input):
        return self._dataset_manager.can_cache(skip_save_processed_input)

    @property
    def data_format(self):
        return self._dataset_manager.data_format



# coding=utf-8
# Copyright (c) 2021 Linuf Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
from tensorflow.keras.layers import Layer, Dense, Dropout, LayerNormalization, \
    Conv2D, GlobalAveragePooling1D

from ludwig.modules.activation_modules import gelu


class MLP(Layer):
    def __init__(self, hidden_size, dropout=0.0):
        super().__init__()
        self.hidden_size = hidden_size
        self.dropout = dropout

    def build(self, shape):
        self.dense1 = Dense(self.hidden_size)
        self.dense2 = Dense(shape[-1])
        self.dropout1 = Dropout(self.dropout)
        self.dropout2 = Dropout(self.dropout)

    def call(self, inputs, **kwargs):
        hidden = self.dropout1(gelu(self.dense1(inputs)))
        return self.dropout2(self.dense2(hidden))


class MixerBlock(Layer):
    def __init__(self, token_dim, channel_dim, dropout=0.0):
        super().__init__()
        self.mlp1 = MLP(token_dim, dropout)
        self.mlp2 = MLP(channel_dim, dropout)
        self.layernorm1 = LayerNormalization()
        self.layernorm2 = LayerNormalization()

    def call(self, inputs, **kwargs):
        hidden = inputs
        hidden = self.layernorm1(hidden)
        hidden = tf.transpose(hidden, [0, 2, 1])
        hidden = self.mlp1(hidden)
        hidden = tf.transpose(hidden, [0, 2, 1])
        mid = hidden + inputs
        hidden = self.layernorm2(mid)
        hidden = self.mlp2(hidden)
        return hidden + mid


class MLPMixer(Layer):
    """MLPMixer

    Implements
    An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale
    https://arxiv.org/abs/2010.11929
    """

    def __init__(
            self,
            patch_size=16,
            embed_size=512,
            token_size=2048,
            channel_dim=256,
            num_layers=8,
            dropout=0.0,
            avg_pool=True,
    ):
        super().__init__()
        self.patch_conv = Conv2D(embed_size, patch_size, patch_size)
        self.mixer_blocks = [
            MixerBlock(token_size, channel_dim, dropout)
            for _ in range(num_layers)
        ]
        self.layer_norm = LayerNormalization()

        self.gap = None
        if avg_pool:
            self.avg_pool = GlobalAveragePooling1D()

    def call(self, inputs, **kwargs):
        batch_size = tf.shape(inputs)[0]
        hidden = inputs
        hidden = self.patch_conv(hidden)
        hidden = tf.reshape(hidden, [batch_size, -1, hidden.shape[-1]])
        for mixer_block in self.mixer_blocks:
            hidden = mixer_block(hidden)
        hidden = self.layer_norm(hidden)
        if self.avg_pool:
            hidden = self.avg_pool(hidden)
        return hidden



"""TF record dataset."""
import contextlib
import math

import os
from distutils.version import LooseVersion

import tensorflow as tf
from ludwig.constants import NAME, PROC_COLUMN
from ludwig.data.batcher.iterable import IterableBatcher
from ludwig.data.dataset.base import Dataset
from ludwig.data.dataset.pandas import PandasDataset
from ludwig.data.dataset.partitioned import PartitionedDataset
from ludwig.utils.data_utils import DATA_TRAIN_HDF5_FP
from ludwig.utils.fs_utils import to_url
from ludwig.utils.misc_utils import get_combined_features, get_proc_features
from ludwig.utils.data_utils import load_json


if LooseVersion(tf.__version__) >= LooseVersion('2.4'):
    AUTOTUNE = tf.data.AUTOTUNE
else:
    AUTOTUNE = tf.data.experimental.AUTOTUNE


class TFRecordDataset(Dataset):
    def __init__(self, url, features, training_set_metadata):
        self.url = to_url(url)
        self.features = [feature[PROC_COLUMN] for feature in features]
        self.training_set_metadata = training_set_metadata

        meta = load_json(os.path.join(self.url, "meta.json"))
        self.size = meta["size"]
        self.compression_type = meta["compression_type"]
        self.reshape_features = {
            feature[PROC_COLUMN]: list((-1, *training_set_metadata[feature[NAME]]['reshape']))
            for feature in features
            if "reshape" in training_set_metadata[feature[NAME]]
        }

    def get(self, feature_name, sample):
        t = sample[feature_name]
        reshape_dim = self.reshape_features.get(feature_name)
        if reshape_dim is not None:
            # When we read a 1D array from disk, we need to reshape it back to its
            # full dimensions.
            t = tf.reshape(t, reshape_dim)
        return t

    def __len__(self):
        return self.size

    @contextlib.contextmanager
    def initialize_batcher(self,
                           batch_size=128,
                           should_shuffle=True,
                           shuffle_buffer_size=None,
                           seed=0,
                           ignore_last=False,
                           horovod=None):
        cur_shard, shard_count = None, None
        if horovod:
            cur_shard, shard_count = horovod.rank(), horovod.size()
        total_samples = self.size
        local_samples = int(total_samples / shard_count) if shard_count else total_samples

        # Below are routine optimizations for tf.dataset
        compression_ext = '.gz' if self.compression_type else ''
        path = os.path.join(self.url, "*.tfrecords{}".format(compression_ext))

        # Now construct the tfrecrd dataset
        files = tf.data.Dataset.list_files(path)

        # interleave the tfrecord files for parallel reading
        dataset = files.interleave(
            lambda x: tf.data.TFRecordDataset(x, compression_type=self.compression_type),
            num_parallel_calls=AUTOTUNE
        )
        # Fetch one element so to get the parser.
        features, feature_lists = self._detect_schema(dataset)
        parser = self._get_parser(features, feature_lists)

        # sharding
        if shard_count and shard_count > 1:
            dataset = dataset.shard(shard_count, cur_shard)

        # parallelize parser
        dataset = dataset.map(parser, num_parallel_calls=AUTOTUNE)
        # Note(Hao) batching: ideally we should put this line before the above map.
        # but the parser func is not vectorized for now.
        dataset = dataset.batch(batch_size)

        # cache
        dataset = dataset.cache()
        if should_shuffle:
            buffer_size = shuffle_buffer_size or local_samples
            dataset = dataset.shuffle(buffer_size)
        dataset = dataset.repeat()
        dataset = dataset.prefetch(AUTOTUNE)

        steps_per_epoch = math.ceil(local_samples / batch_size)
        batcher = IterableBatcher(self,
                                  dataset,
                                  steps_per_epoch,
                                  ignore_last=ignore_last)
        yield batcher

    def _detect_schema(self, dataset):
        features = {}
        feature_lists = {}

        serialized = next(iter(dataset.map(lambda serialized: serialized)))
        seq_ex = tf.train.SequenceExample.FromString(serialized.numpy())

        def _get_feature_type(feature=None, type_=None):
            if type_:
                return {
                    int: tf.int64,
                    float: tf.float32,
                    str: tf.string,
                    bytes: tf.string,
                }[type_]

            if feature:
                if feature.HasField('int64_list'):
                    return tf.int64
                if feature.HasField('float_list'):
                    return tf.float32
                if feature.HasField('bytes_list'):
                    return tf.string

        if seq_ex.context.feature:
            for key, feature in seq_ex.context.feature.items():
                features[key] = tf.io.FixedLenFeature(
                    (), _get_feature_type(feature=feature))

        if seq_ex.feature_lists.feature_list:
            for key, feature_list in seq_ex.feature_lists.feature_list.items():
                feature_lists[key] = tf.io.FixedLenSequenceFeature(
                    (), _get_feature_type(feature=feature_list.feature[0]))
        return features, feature_lists

    def _get_parser(self, features, feature_lists):
        def parse_example(feats):

            def parse(binary):
                return tf.io.parse_single_example(binary, features=feats)

            return parse

        def parse_sequence(feats, feat_lists):

            def parse(binary):
                context, sequence = tf.io.parse_single_sequence_example(
                    binary,
                    context_features=feats,
                    sequence_features=feat_lists
                )
                context.update(sequence)
                return context

            return parse

        if feature_lists:
            return parse_sequence(features, feature_lists)
        else:
            return parse_example(features)


class TFRecordDatasetManager(object):
    def __init__(self, backend):
        self.backend = backend

    def create(self, dataset, config, training_set_metadata):
        """Create a TFRecordDataset.

        Args:
            dataset (str): urls to the tfrecord binary files.
            config ():
            training_set_metadata ():

        Returns:
            TFRecordDataset
        """
        features = get_combined_features(config)
        return TFRecordDataset(
            dataset,
            features,
            training_set_metadata
        )

    def create_inference_dataset(self, dataset, tag, config, training_set_metadata):
        """We don't use TFRecord for inference."""
        if self.backend.df_engine.partitioned:
            return PartitionedDataset(
                dataset,
                get_proc_features(config),
                training_set_metadata.get(DATA_TRAIN_HDF5_FP)
            )
        else:
            return PandasDataset(
                dataset,
                get_proc_features(config),
                training_set_metadata.get(DATA_TRAIN_HDF5_FP)
            )

    def save(self, cache_path, dataset, config, training_set_metadata, tag):
        dataset_tfrecord_fp = cache_path

        features = get_combined_features(config)
        for feature in features:
            name = feature[NAME]
            proc_column = feature[PROC_COLUMN]
            reshape = training_set_metadata[name].get('reshape')
            if reshape:
                dataset[proc_column] = self.backend.df_engine.map_objects(
                    dataset[proc_column],
                    lambda x: x.reshape(-1))
        self.backend.df_engine.to_tfrecord(dataset, dataset_tfrecord_fp)
        return dataset_tfrecord_fp

    def can_cache(self, skip_save_processed_input):
        return self.backend.is_coordinator()

    @property
    def data_format(self):
        return 'tfrecords'


def get_compression_ext(compression_type):
    return '.gz' if compression_type else ''


def get_part_filename(i, compression_ext):
    return f"part.{str(i).zfill(5)}.tfrecords{compression_ext}"



import logging
import os
import shutil

import numpy as np

from ludwig import __file__ as ludwig_path
from ludwig.api import LudwigModel
from ludwig.backend import LOCAL_BACKEND, LocalBackend
from ludwig.constants import (BINARY, CATEGORY, NUMERICAL, PREDICTIONS,
                              PROBABILITIES, PROBABILITY, SEQUENCE, SET, TEXT,
                              TYPE, VECTOR, NAME)
from ludwig.globals import (MODEL_HYPERPARAMETERS_FILE_NAME,
                            MODEL_WEIGHTS_FILE_NAME,
                            TRAIN_SET_METADATA_FILE_NAME)
from ludwig.utils.data_utils import load_json

logger = logging.getLogger(__name__)


class LudwigNeuropodModelWrapper:
    def __init__(self, data_root, backend):
        self.ludwig_model = LudwigModel.load(data_root, backend=backend)

    def __call__(self, **kwargs):
        data_dict = kwargs
        for key in data_dict:
            data_dict[key] = np.squeeze(data_dict[key], axis=1)
        predicted, _ = self.ludwig_model.predict(
            dataset=data_dict,
            return_type='dict'
        )
        # print(predicted, file=sys.stderr)
        return postprocess_for_neuropod(
            predicted, self.ludwig_model.config
        )


def get_model(data_root):
    return LudwigNeuropodModelWrapper(data_root, backend=LOCAL_BACKEND)


def get_test_model(data_root):
    class LocalTestBackend(LocalBackend):
        @property
        def supports_multiprocessing(self):
            return False
    return LudwigNeuropodModelWrapper(data_root, backend=LocalTestBackend())


def postprocess_for_neuropod(predicted, config):
    postprocessed = {}
    for output_feature in config['output_features']:
        feature_name = output_feature[NAME]
        feature_type = output_feature[TYPE]
        if feature_type == BINARY:
            postprocessed[feature_name + "_predictions"] = \
                np.expand_dims(
                    predicted[feature_name][PREDICTIONS].astype('str'), 1)
            postprocessed[feature_name + "_probability"] = \
                np.expand_dims(
                    predicted[feature_name][PROBABILITY].astype('float64'),
                    1)
            postprocessed[feature_name + "_probabilities"] = \
                predicted[feature_name][PROBABILITIES].astype('float64')
        elif feature_type == NUMERICAL:
            postprocessed[feature_name + "_predictions"] = \
                np.expand_dims(
                    predicted[feature_name][PREDICTIONS].astype('float64'),
                    1)
        elif feature_type == CATEGORY:
            postprocessed[feature_name + "_predictions"] = np.expand_dims(
                np.array(predicted[feature_name][PREDICTIONS], dtype='str'), 1
            )
            postprocessed[feature_name + "_probability"] = \
                np.expand_dims(
                    predicted[feature_name][PROBABILITY].astype('float64'),
                    1)
            postprocessed[feature_name + "_probabilities"] = \
                predicted[feature_name][PROBABILITIES].astype('float64')
        elif feature_type == SEQUENCE or feature_type == TEXT:
            predictions = list(map(
                lambda x: ' '.join(x),
                predicted[feature_name][PREDICTIONS]
            ))
            postprocessed[feature_name + "_predictions"] = np.expand_dims(
                np.array(predictions, dtype='str'), 1
            )
        elif feature_type == SET:
            predictions = list(map(
                lambda x: ' '.join(x),
                predicted[feature_name][PREDICTIONS]
            ))
            postprocessed[feature_name + "_predictions"] = np.expand_dims(
                np.array(predictions, dtype='str'), 1
            )
            probability = list(map(
                lambda x: ' '.join([str(e) for e in x]),
                predicted[feature_name]['probability']
            ))
            postprocessed[feature_name + "_probability"] = np.expand_dims(
                np.array(probability, dtype='str'), 1
            )
            postprocessed[feature_name + "_probabilities"] = \
                predicted[feature_name][PROBABILITIES].astype('float64')
        elif feature_type == VECTOR:
            postprocessed[feature_name + "_predictions"] = \
                predicted[feature_name][PREDICTIONS].astype('float64')
        else:
            postprocessed[feature_name + "_predictions"] = np.expand_dims(
                np.array(predicted[feature_name][PREDICTIONS], dtype='str'), 1
            )
    return postprocessed


def export_neuropod(
        ludwig_model_path,
        neuropod_path,
        neuropod_model_name="ludwig_model",
        entrypoint='get_model',
):
    try:
        from neuropod.backends.python.packager import create_python_neuropod
    except ImportError:
        raise RuntimeError(
            'The "neuropod" package is not installed in your environment.'
        )

    data_paths = [
        {
            "path": os.path.join(
                ludwig_model_path, MODEL_HYPERPARAMETERS_FILE_NAME
            ),
            "packaged_name": MODEL_HYPERPARAMETERS_FILE_NAME
        },
        {
            "path": os.path.join(
                ludwig_model_path, TRAIN_SET_METADATA_FILE_NAME
            ),
            "packaged_name": TRAIN_SET_METADATA_FILE_NAME
        },
        {
            "path": os.path.join(
                ludwig_model_path, 'checkpoint'
            ),
            "packaged_name": 'checkpoint'
        },
    ]
    for filename in os.listdir(ludwig_model_path):
        if MODEL_WEIGHTS_FILE_NAME in filename:
            data_paths.append(
                {
                    "path": os.path.join(
                        ludwig_model_path, filename
                    ),
                    "packaged_name": filename
                }
            )

    logger.debug('data_paths: {}'.format(data_paths))

    ludwig_config = load_json(
        os.path.join(
            ludwig_model_path,
            MODEL_HYPERPARAMETERS_FILE_NAME
        )
    )
    training_set_metadata = load_json(
        os.path.join(
            ludwig_model_path,
            TRAIN_SET_METADATA_FILE_NAME
        )
    )

    input_spec = []
    for feature in ludwig_config['input_features']:
        input_spec.append({
            NAME: feature[NAME],
            "dtype": "str",
            "shape": (None, 1)
        })
    logger.debug('input_spec: {}'.format(input_spec))

    output_spec = []
    for feature in ludwig_config['output_features']:
        feature_type = feature[TYPE]
        feature_name = feature[NAME]
        if feature_type == BINARY:
            output_spec.append({
                "name": feature_name + '_predictions',
                "dtype": "str",
                "shape": (None, 1)
            })
            output_spec.append({
                "name": feature_name + '_probability',
                "dtype": "float64",
                "shape": (None, 1)
            })
            output_spec.append({
                "name": feature_name + '_probabilities',
                "dtype": "float64",
                "shape": (None, 2)
            })
        elif feature_type == NUMERICAL:
            output_spec.append({
                "name": feature_name + '_predictions',
                "dtype": "float64",
                "shape": (None, 1)
            })
        elif feature_type == CATEGORY:
            output_spec.append({
                "name": feature_name + '_predictions',
                "dtype": "str",
                "shape": (None, 1)
            })
            output_spec.append({
                "name": feature_name + '_probability',
                "dtype": "float64",
                "shape": (None, 1)
            })
            output_spec.append({
                "name": feature_name + '_probabilities',
                "dtype": "float64",
                "shape": (
                    None,
                    training_set_metadata[feature[NAME]]['vocab_size']
                )
            })
        elif feature_type == SEQUENCE or feature_type == TEXT:
            output_spec.append({
                "name": feature_name + '_predictions',
                "dtype": "str",
                "shape": (None, 1)
            })
        elif feature_type == SET:
            output_spec.append({
                "name": feature_name + '_predictions',
                "dtype": "str",
                "shape": (None, 1)
            })
            output_spec.append({
                "name": feature_name + '_probability',
                "dtype": "str",
                "shape": (None, 1)
            })
            output_spec.append({
                "name": feature_name + '_probabilities',
                "dtype": "float64",
                "shape": (
                    None,
                    training_set_metadata[feature[NAME]]['vocab_size']
                )
            })
        elif feature_type == VECTOR:
            output_spec.append({
                "name": feature['name'] + '_predictions',
                "dtype": "float64",
                "shape": (
                    None,
                    training_set_metadata[feature[NAME]]['vector_size']
                )
            })
        else:
            output_spec.append({
                "name": feature['name'] + '_predictions',
                "dtype": "str",
                "shape": (None, 1)
            })
    logger.debug('output_spec: {}'.format(output_spec))

    if os.path.exists(neuropod_path):
        if os.path.isfile(neuropod_path):
            logger.warning('Removing file: {}'.format(neuropod_path))
            os.remove(neuropod_path)
        else:
            logger.warning('Removing directory: {}'.format(neuropod_path))
            shutil.rmtree(neuropod_path, ignore_errors=True)

    from pathlib import Path
    path = Path(ludwig_path)
    logger.debug('python_root: {}'.format(path.parent.parent))

    create_python_neuropod(
        neuropod_path=neuropod_path,
        model_name=neuropod_model_name,
        data_paths=data_paths,
        code_path_spec=[{
            "python_root": path.parent.parent,
            "dirs_to_package": [
                "ludwig"  # Package everything in the python_root
            ],
        }],
        entrypoint_package="ludwig.utils.neuropod_utils",
        entrypoint=entrypoint,
        skip_virtualenv=True,
        input_spec=input_spec,
        output_spec=output_spec
    )
    logger.info('Neuropod saved to: {}'.format(neuropod_path))



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
import shutil

import pandas as pd
import pytest

from ludwig.constants import NAME
from ludwig.experiment import experiment_cli

from tests.integration_tests.utils import binary_feature, sequence_feature, \
    set_feature, text_feature, vector_feature
from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import numerical_feature

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)


def run_experiment(input_features, output_features, **kwargs):
    """
    Helper method to avoid code repetition in running an experiment. Deletes
    the data saved to disk after running the experiment
    :param input_features: list of input feature dictionaries
    :param output_features: list of output feature dictionaries
    **kwargs you may also pass extra parameters to the experiment as keyword
    arguments
    :return: None
    """
    config = None
    if input_features is not None and output_features is not None:
        # This if is necessary so that the caller can call with
        # config_file (and not config)
        config = {
            'input_features': input_features,
            'output_features': output_features,
            'combiner': {
                'type': 'concat',
                'fc_size': 64,
                'num_fc_layers': 5
            },
            'training': {'epochs': 2}
        }

    args = {
        'config': config,
        'skip_save_processed_input': True,
        'skip_save_progress': True,
        'skip_save_unprocessed_output': True,
        'skip_save_model': True,
        'skip_save_log': True
    }
    args.update(kwargs)

    exp_dir_name = experiment_cli(**args)
    shutil.rmtree(exp_dir_name, ignore_errors=True)


@pytest.mark.parametrize(
    'input_test_feature, output_test_feature, output_loss_parameter',
    [
        # numerical features
        (numerical_feature(), numerical_feature(), None),
        (
                numerical_feature(normalization='minmax'),
                numerical_feature(),
                {'loss': {'type': 'mean_squared_error'}}
        ),
        (
                numerical_feature(normalization='zscore'),
                numerical_feature(),
                {'loss': {'type': 'mean_absolute_error'}}
        ),

        # binary feature
        (binary_feature(), binary_feature(), None),

        # Categorical feature
        (category_feature(), category_feature(), None),
        (
                category_feature(),
                category_feature(),
                {'loss': {'type': 'softmax_cross_entropy'}}
        ),
        (
                category_feature(),
                category_feature(),
                {'loss': {
                    'type': 'sampled_softmax_cross_entropy',
                    'sampler': 'fixed_unigram',
                    'negative_samples': 10
                }
                }
        ),
        (
                category_feature(),
                category_feature(),
                {'loss': {
                    'type': 'sampled_softmax_cross_entropy',
                    'sampler': 'uniform',
                    'negative_samples': 10
                }
                }
        ),
        (
                category_feature(),
                category_feature(),
                {'loss': {
                    'type': 'sampled_softmax_cross_entropy',
                    'sampler': 'log_uniform',
                    'negative_samples': 10
                }
                }
        ),
        (
                category_feature(),
                category_feature(),
                {'loss': {
                    'type': 'sampled_softmax_cross_entropy',
                    'sampler': 'learned_unigram',
                    'negative_samples': 10
                }
                }
        )
    ]
)
def test_feature(input_test_feature, output_test_feature,
                 output_loss_parameter, csv_filename):
    input_features = [
        input_test_feature
    ]

    of_test_feature = output_test_feature
    if output_loss_parameter is not None:
        of_test_feature.update(output_loss_parameter)
    output_features = [of_test_feature]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename,
                             1001)

    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize(
    'input_test_feature, output_test_feature',
    [
        ([category_feature()],
         [binary_feature(), binary_feature()]),
        ([category_feature()],
         [category_feature(vocab_size=5), category_feature(vocab_size=7)]),
        ([category_feature()],
         [numerical_feature(), numerical_feature()]),
        ([category_feature()],
         [sequence_feature(vocab_size=5), sequence_feature(vocab_size=7)]),
        ([category_feature()],
         [set_feature(vocab_size=5), set_feature(vocab_size=7)]),
        ([category_feature()],
         [text_feature(vocab_size=5), text_feature(vocab_size=7)]),
        ([category_feature()],
         [vector_feature(), vector_feature()]),
    ]
)
def test_feature_multiple_outputs(input_test_feature, output_test_feature,
                                  csv_filename):
    # Generate test data
    rel_path = generate_data(input_test_feature, output_test_feature,
                             csv_filename, 1001)

    run_experiment(input_test_feature, output_test_feature, dataset=rel_path)


def test_category_int_dtype(tmpdir):
    feature = category_feature()
    input_features = [feature]
    output_features = [binary_feature()]

    csv_fname = generate_data(input_features, output_features,
                              os.path.join(tmpdir, 'dataset.csv'))
    df = pd.read_csv(csv_fname)

    distinct_values = df[feature[NAME]].drop_duplicates().values
    value_map = {v: idx for idx, v in enumerate(distinct_values)}
    df[feature[NAME]] = df[feature[NAME]].map(
        lambda x: value_map[x]
    )

    run_experiment(input_features, output_features, dataset=df)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
import shutil
import sys
import uuid
from collections import namedtuple

import pandas as pd
import pytest
import yaml

from skimage.io import imread

from ludwig.api import LudwigModel
from ludwig.backend import LOCAL_BACKEND
from ludwig.data.concatenate_datasets import concatenate_df
from ludwig.data.preprocessing import preprocess_for_training
from ludwig.experiment import experiment_cli
from ludwig.features.h3_feature import H3InputFeature
from ludwig.predict import predict_cli
from ludwig.utils.data_utils import read_csv
from ludwig.utils.defaults import default_random_seed

from tests.conftest import delete_temporary_data
from tests.integration_tests.utils import ENCODERS, HF_ENCODERS, \
    HF_ENCODERS_SHORT, slow, create_data_set_to_use
from tests.integration_tests.utils import audio_feature
from tests.integration_tests.utils import bag_feature
from tests.integration_tests.utils import binary_feature
from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import date_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import \
    generate_output_features_with_dependencies
from tests.integration_tests.utils import h3_feature
from tests.integration_tests.utils import image_feature
from tests.integration_tests.utils import numerical_feature
from tests.integration_tests.utils import run_experiment
from tests.integration_tests.utils import sequence_feature
from tests.integration_tests.utils import set_feature
from tests.integration_tests.utils import spawn
from tests.integration_tests.utils import text_feature
from tests.integration_tests.utils import timeseries_feature
from tests.integration_tests.utils import vector_feature
from tests.integration_tests.utils import LocalTestBackend


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)


@pytest.mark.parametrize('encoder', ENCODERS)
def test_experiment_text_feature_non_HF(encoder, csv_filename):
    input_features = [
        text_feature(
            vocab_size=30,
            min_len=1,
            encoder=encoder,
            preprocessing={'word_tokenizer': 'space'}
        )
    ]
    output_features = [category_feature(vocab_size=2)]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@spawn
def run_experiment_with_encoder(encoder, csv_filename):
    # Run in a subprocess to clear TF and prevent OOM
    # This also allows us to use GPU resources
    input_features = [
        text_feature(
            vocab_size=30,
            min_len=1,
            encoder=encoder,
        )
    ]
    output_features = [category_feature(vocab_size=2)]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.distributed
@pytest.mark.parametrize('encoder', HF_ENCODERS_SHORT)
def test_experiment_text_feature_HF(encoder, csv_filename):
    run_experiment_with_encoder(encoder, csv_filename)


@slow
@pytest.mark.parametrize('encoder', HF_ENCODERS)
def test_experiment_text_feature_HF_full(encoder, csv_filename):
    run_experiment_with_encoder(encoder, csv_filename)


def test_experiment_seq_seq(csv_filename):
    # Single Sequence input, single sequence output
    # Only the following encoders are working
    input_features = [text_feature(reduce_output=None, encoder='rnn')]
    output_features = [text_feature(reduce_input=None, decoder='tagger')]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    encoders2 = ['cnnrnn', 'stacked_cnn']
    for encoder in encoders2:
        logger.info('seq to seq test, Encoder: {0}'.format(encoder))
        input_features[0]['encoder'] = encoder
        run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_seq_seq_model_def_file(csv_filename, yaml_filename):
    # seq-to-seq test to use config file instead of dictionary
    input_features = [text_feature(reduce_output=None, encoder='embed')]
    output_features = [
        text_feature(reduce_input=None, vocab_size=3, decoder='tagger')
    ]

    # Save the config to a yaml file
    config = {
        'input_features': input_features,
        'output_features': output_features,
        'combiner': {'type': 'concat', 'fc_size': 14},
        'training': {'epochs': 2}
    }
    with open(yaml_filename, 'w') as yaml_out:
        yaml.safe_dump(config, yaml_out)

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(
        None, None, dataset=rel_path, config=yaml_filename
    )


def test_experiment_seq_seq_train_test_valid(csv_filename):
    # seq-to-seq test to use train, test, validation files
    input_features = [text_feature(reduce_output=None, encoder='rnn')]
    output_features = [
        text_feature(reduce_input=None, vocab_size=3, decoder='tagger')
    ]

    train_csv = generate_data(
        input_features, output_features, 'tr_' + csv_filename
    )
    test_csv = generate_data(
        input_features, output_features, 'test_' + csv_filename, 20
    )
    valdation_csv = generate_data(
        input_features, output_features, 'val_' + csv_filename, 20
    )

    run_experiment(
        input_features,
        output_features,
        training_set=train_csv,
        test_set=test_csv,
        validation_set=valdation_csv
    )

    input_features[0]['encoder'] = 'parallel_cnn'
    # Save intermediate output
    run_experiment(
        input_features,
        output_features,
        training_set=train_csv,
        test_set=test_csv,
        validation_set=valdation_csv
    )

    # Delete the temporary data created
    # This test is saving the processed data to hdf5
    for prefix in ['tr_', 'test_', 'val_']:
        if os.path.isfile(prefix + csv_filename):
            os.remove(prefix + csv_filename)


def test_experiment_multi_input_intent_classification(csv_filename):
    # Multiple inputs, Single category output
    input_features = [
        text_feature(vocab_size=10, min_len=1, representation='sparse'),
        category_feature(vocab_size=10)
    ]
    output_features = [category_feature(vocab_size=2, reduce_input='sum')]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    for encoder in ENCODERS:
        input_features[0]['encoder'] = encoder
        run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_multiclass_with_class_weights(csv_filename):
    # Multiple inputs, Single category output
    input_features = [category_feature(vocab_size=10)]
    output_features = [category_feature(vocab_size=3,
                                        loss={"class_weights": [0, 1, 2, 3]})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_multilabel_with_class_weights(csv_filename):
    # Multiple inputs, Single category output
    input_features = [category_feature(vocab_size=10)]
    output_features = [set_feature(vocab_size=3,
                                   loss={"class_weights": [0, 0, 1, 2, 3]}
                                   )]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize(
    'output_features',
    [
        # baseline test case
        [
            category_feature(vocab_size=2, reduce_input='sum'),
            sequence_feature(vocab_size=10, max_len=5),
            numerical_feature()
        ],

        # use generator as decoder
        [
            category_feature(vocab_size=2, reduce_input='sum'),
            sequence_feature(vocab_size=10, max_len=5, decoder='generator'),
            numerical_feature()
        ],

        # Generator decoder and reduce_input = None
        [
            category_feature(vocab_size=2, reduce_input='sum'),
            sequence_feature(max_len=5, decoder='generator',
                             reduce_input=None),
            numerical_feature(normalization='minmax')
        ],

        # output features with dependencies single dependency
        generate_output_features_with_dependencies('feat3', ['feat1']),

        # output features with dependencies multiple dependencies
        generate_output_features_with_dependencies('feat3',
                                                   ['feat1', 'feat2']),

        # output features with dependencies multiple dependencies
        generate_output_features_with_dependencies('feat2',
                                                   ['feat1', 'feat3']),

        # output features with dependencies
        generate_output_features_with_dependencies('feat1', ['feat2'])
    ]
)
def test_experiment_multiple_seq_seq(csv_filename, output_features):
    input_features = [
        text_feature(vocab_size=100, min_len=1, encoder='stacked_cnn'),
        numerical_feature(normalization='zscore'),
        category_feature(vocab_size=10, embedding_size=5),
        set_feature(),
        sequence_feature(vocab_size=10, max_len=10, encoder='embed')
    ]
    output_features = output_features

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize('skip_save_processed_input', [True, False])
@pytest.mark.parametrize('in_memory', [True, False])
@pytest.mark.parametrize('image_source', ['file', 'ndarray'])
@pytest.mark.parametrize('num_channels', [1, 3])
def test_basic_image_feature(num_channels, image_source, in_memory,
                             skip_save_processed_input, csv_filename):
    # Image Inputs
    image_dest_folder = os.path.join(os.getcwd(), 'generated_images')

    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder='stacked_cnn',
            preprocessing={
                'in_memory': in_memory,
                'height': 12,
                'width': 12,
                'num_channels': num_channels,
                'num_processes': 5
            },
            fc_size=16,
            num_filters=8
        )
    ]
    output_features = [
        category_feature(vocab_size=2, reduce_input='sum')
    ]

    rel_path = generate_data(input_features, output_features, csv_filename)

    if image_source == 'file':
        # use images from file
        run_experiment(
            input_features,
            output_features,
            dataset=rel_path,
            skip_save_processed_input=skip_save_processed_input
        )
    else:
        # import image from file and store in dataframe as ndarrays
        df = pd.read_csv(rel_path)
        image_feature_name = input_features[0]['name']
        df[image_feature_name] = df[image_feature_name].apply(
            lambda x: imread(x))

        run_experiment(
            input_features,
            output_features,
            dataset=df,
            skip_save_processed_input=skip_save_processed_input
        )

    # Delete the temporary data created
    shutil.rmtree(image_dest_folder, ignore_errors=True)


ImageParms = namedtuple(
    'ImageTestParms',
    'image_encoder in_memory_flag skip_save_processed_input'
)


@pytest.mark.parametrize(
    'image_parms',
    [
        ImageParms('resnet', True, True),
        ImageParms('stacked_cnn', True, True),
        ImageParms('stacked_cnn', False, False)
    ]
)
def test_experiment_image_inputs(image_parms: ImageParms, csv_filename: str):
    # Image Inputs
    image_dest_folder = os.path.join(os.getcwd(), 'generated_images')

    # Resnet encoder
    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder='resnet',
            preprocessing={
                'in_memory': True,
                'height': 12,
                'width': 12,
                'num_channels': 3,
                'num_processes': 5
            },
            fc_size=16,
            num_filters=8
        ),
        text_feature(encoder='embed', min_len=1),
        numerical_feature(normalization='zscore')
    ]
    output_features = [
        category_feature(vocab_size=2, reduce_input='sum'),
        numerical_feature()
    ]

    input_features[0]['encoder'] = image_parms.image_encoder
    input_features[0]['preprocessing'][
        'in_memory'] = image_parms.in_memory_flag
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(
        input_features,
        output_features,
        dataset=rel_path,
        skip_save_processed_input=image_parms.skip_save_processed_input,
    )

    # Delete the temporary data created
    shutil.rmtree(image_dest_folder)


IMAGE_DATA_FORMATS_TO_TEST = ['csv', 'df', 'hdf5']

@pytest.mark.parametrize('test_in_memory', [True, False])
@pytest.mark.parametrize('test_format', IMAGE_DATA_FORMATS_TO_TEST)
@pytest.mark.parametrize('train_in_memory', [True, False])
@pytest.mark.parametrize('train_format', IMAGE_DATA_FORMATS_TO_TEST)
def test_experiment_image_dataset(
        train_format, train_in_memory,
        test_format, test_in_memory
):
    # primary focus of this test is to determine if exceptions are
    # raised for different data set formats and in_memory setting
    # Image Inputs
    image_dest_folder = os.path.join(os.getcwd(), 'generated_images')

    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder='stacked_cnn',
            preprocessing={
                'in_memory': True,
                'height': 12,
                'width': 12,
                'num_channels': 3,
                'num_processes': 5
            },
            fc_size=16,
            num_filters=8
        ),
    ]
    output_features = [
        category_feature(vocab_size=2, reduce_input='sum'),
    ]

    config = {
        'input_features': input_features,
        'output_features': output_features,
        'combiner': {
            'type': 'concat',
            'fc_size': 14
        },
        'preprocessing': {},
        'training': {'epochs': 2}
    }

    # create temporary name for train and test data sets
    train_csv_filename = 'train_' + uuid.uuid4().hex[:10].upper() + '.csv'
    test_csv_filename = 'test_' + uuid.uuid4().hex[:10].upper() + '.csv'

    # setup training data format to test
    train_data = generate_data(input_features, output_features,
                               train_csv_filename)
    config['input_features'][0]['preprocessing']['in_memory'] \
        = train_in_memory
    training_set_metadata = None

    backend = LocalTestBackend()
    if train_format == 'hdf5':
        # hdf5 format
        train_set, _, _, training_set_metadata = preprocess_for_training(
            config,
            dataset=train_data,
            backend=backend,
        )
        train_dataset_to_use = train_set.data_hdf5_fp
    else:
        train_dataset_to_use = create_data_set_to_use(train_format, train_data)

    # define Ludwig model
    model = LudwigModel(
        config=config,
        backend=backend,
    )
    model.train(
        dataset=train_dataset_to_use,
        training_set_metadata=training_set_metadata
    )

    model.config['input_features'][0]['preprocessing']['in_memory'] \
        = test_in_memory

    # setup test data format to test
    test_data = generate_data(input_features, output_features,
                              test_csv_filename)

    if test_format == 'hdf5':
        # hdf5 format
        # create hdf5 data set
        _, test_set, _, training_set_metadata_for_test = preprocess_for_training(
            model.config,
            dataset=test_data,
            backend=backend,
        )
        test_dataset_to_use = test_set.data_hdf5_fp
    else:
        test_dataset_to_use = create_data_set_to_use(test_format, test_data)

    # run functions with the specified data format
    model.evaluate(dataset=test_dataset_to_use)
    model.predict(dataset=test_dataset_to_use)

    # Delete the temporary data created
    shutil.rmtree(image_dest_folder)
    delete_temporary_data(train_csv_filename)
    delete_temporary_data(test_csv_filename)


DATA_FORMATS_TO_TEST = [
    'csv', 'df', 'dict', 'excel', 'excel_xls', 'feather', 'fwf', 'hdf5', 'html',
    'json', 'jsonl', 'parquet', 'pickle', 'stata', 'tsv'
]
@pytest.mark.parametrize('data_format', DATA_FORMATS_TO_TEST)
def test_experiment_dataset_formats(data_format):
    # primary focus of this test is to determine if exceptions are
    # raised for different data set formats and in_memory setting

    input_features = [
        numerical_feature(),
        category_feature()
    ]
    output_features = [
        category_feature(),
        numerical_feature()
    ]

    config = {
        'input_features': input_features,
        'output_features': output_features,
        'combiner': {
            'type': 'concat',
            'fc_size': 14
        },
        'preprocessing': {},
        'training': {'epochs': 2}
    }

    # create temporary name for train and test data sets
    csv_filename = 'train_' + uuid.uuid4().hex[:10].upper() + '.csv'

    # setup training data format to test
    raw_data = generate_data(input_features, output_features,
                               csv_filename)

    training_set_metadata = None

    if data_format == 'hdf5':
        # hdf5 format
        training_set, _, _, training_set_metadata = preprocess_for_training(
            config,
            dataset=raw_data
        )
        dataset_to_use = training_set.data_hdf5_fp
    else:
        dataset_to_use = create_data_set_to_use(data_format, raw_data)

    # define Ludwig model
    model = LudwigModel(
        config=config
    )
    model.train(
        dataset=dataset_to_use,
        training_set_metadata=training_set_metadata,
        random_seed=default_random_seed
    )

    # # run functions with the specified data format
    model.evaluate(dataset=dataset_to_use)
    model.predict(dataset=dataset_to_use)

    # Delete the temporary data created
    delete_temporary_data(csv_filename)



def test_experiment_audio_inputs(csv_filename):
    # Audio Inputs
    audio_dest_folder = os.path.join(os.getcwd(), 'generated_audio')

    input_features = [
        audio_feature(
            folder=audio_dest_folder
        )
    ]
    output_features = [
        binary_feature()
    ]

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)

    # Delete the temporary data created
    shutil.rmtree(audio_dest_folder)


def test_experiment_tied_weights(csv_filename):
    # Single sequence input, single category output
    input_features = [
        text_feature(
            name='text_feature1',
            min_len=1,
            encoder='cnnrnn',
            reduce_output='sum'
        ),
        text_feature(
            name='text_feature2',
            min_len=1,
            encoder='cnnrnn',
            reduce_output='sum',
            tied_weights='text_feature1'
        )
    ]
    output_features = [category_feature(vocab_size=2, reduce_input='sum')]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    for encoder in ENCODERS:
        input_features[0]['encoder'] = encoder
        input_features[1]['encoder'] = encoder
        run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize('enc_cell_type', ['lstm', 'rnn', 'gru'])
@pytest.mark.parametrize('attention', [False, True])
def test_sequence_tagger(
        enc_cell_type,
        attention,
        csv_filename
):
    # Define input and output features
    input_features = [
        sequence_feature(
            max_len=10,
            encoder='rnn',
            cell_type=enc_cell_type,
            reduce_output=None
        )
    ]
    output_features = [
        sequence_feature(
            max_len=10,
            decoder='tagger',
            attention=attention,
            reduce_input=None
        )
    ]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    # run the experiment
    run_experiment(input_features, output_features, dataset=rel_path)


def test_sequence_tagger_text(
        csv_filename
):
    # Define input and output features
    input_features = [
        text_feature(
            max_len=10,
            encoder='rnn',
            reduce_output=None
        )
    ]
    output_features = [
        sequence_feature(
            max_len=10,
            decoder='tagger',
            reduce_input=None
        )
    ]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    # run the experiment
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize('sequence_combiner_encoder', ENCODERS[:-2])
def test_experiment_sequence_combiner(sequence_combiner_encoder, csv_filename):
    # Sequence combiner
    input_features = [
        sequence_feature(
            name='seq1',
            min_len=5,
            max_len=5,
            encoder='rnn',
            cell_type='lstm',
            reduce_output=None
        ),
        sequence_feature(
            name='seq2',
            min_len=5,
            max_len=5,
            encoder='rnn',
            cell_type='lstm',
            reduce_output=None
        ),
        category_feature(vocab_size=5)
    ]
    output_features = [
        category_feature(reduce_input='sum', vocab_size=5)
    ]

    config = {
        'input_features': input_features,
        'output_features': output_features,
        'training': {
            'epochs': 2
        },
        'combiner': {
            'type': 'sequence',
            'encoder': 'rnn',
            'main_sequence_feature': 'seq1',
            'reduce_output': None,
        }
    }

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    for encoder in ENCODERS[:-2]:
        logger.error('sequence combiner. encoders: {0}, {1}'.format(
            encoder,
            encoder
        ))
        input_features[0]['encoder'] = encoder
        input_features[1]['encoder'] = encoder

        config['input_features'] = input_features

        exp_dir_name = experiment_cli(
            config,
            skip_save_processed_input=False,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
            dataset=rel_path
        )
        shutil.rmtree(exp_dir_name, ignore_errors=True)


def test_experiment_model_resume(csv_filename):
    # Single sequence input, single category output
    # Tests saving a model file, loading it to rerun training and predict
    input_features = [sequence_feature(encoder='rnn', reduce_output='sum')]
    output_features = [category_feature(vocab_size=2, reduce_input='sum')]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        'input_features': input_features,
        'output_features': output_features,
        'combiner': {'type': 'concat', 'fc_size': 14},
        'training': {'epochs': 2}
    }

    _, _, _, _, output_dir = experiment_cli(config, dataset=rel_path)
    logger.info('Experiment Directory: {0}'.format(output_dir))

    experiment_cli(
        config,
        dataset=rel_path,
        model_resume_path=output_dir
    )

    predict_cli(os.path.join(output_dir, 'model'), dataset=rel_path)
    shutil.rmtree(output_dir, ignore_errors=True)


def test_experiment_various_feature_types(csv_filename):
    input_features = [binary_feature(), bag_feature()]
    output_features = [set_feature(max_len=3, vocab_size=5)]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_timeseries(csv_filename):
    input_features = [timeseries_feature()]
    output_features = [binary_feature()]

    encoders2 = [
        'transformer'
    ]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    for encoder in encoders2:
        input_features[0]['encoder'] = encoder
        run_experiment(input_features, output_features, dataset=rel_path)


def test_visual_question_answering(csv_filename):
    image_dest_folder = os.path.join(os.getcwd(), 'generated_images')
    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder='resnet',
            preprocessing={
                'in_memory': True,
                'height': 8,
                'width': 8,
                'num_channels': 3,
                'num_processes': 5
            },
            fc_size=8,
            num_filters=8
        ),
        text_feature(encoder='embed', min_len=1, level='word'),
    ]
    output_features = [sequence_feature(decoder='generator', cell_type='lstm')]
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)

    # Delete the temporary data created
    shutil.rmtree(image_dest_folder)


def test_image_resizing_num_channel_handling(csv_filename):
    """
    This test creates two image datasets with 3 channels and 1 channel. The
    combination of this data is used to train a model. This checks the cases
    where the user may or may not specify a number of channels in the
    config
    :param csv_filename:
    :return:
    """
    # Image Inputs
    image_dest_folder = os.path.join(os.getcwd(), 'generated_images')

    # Resnet encoder
    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder='resnet',
            preprocessing={
                'in_memory': True,
                'height': 8,
                'width': 8,
                'num_channels': 3,
                'num_processes': 5
            },
            fc_size=8,
            num_filters=8
        ),
        text_feature(encoder='embed', min_len=1),
        numerical_feature(normalization='minmax')
    ]
    output_features = [binary_feature(), numerical_feature()]
    rel_path = generate_data(
        input_features, output_features, csv_filename, num_examples=50
    )

    df1 = read_csv(rel_path)

    input_features[0]['preprocessing']['num_channels'] = 1
    rel_path = generate_data(
        input_features, output_features, csv_filename, num_examples=50
    )
    df2 = read_csv(rel_path)

    df = concatenate_df(df1, df2, None, LOCAL_BACKEND)
    df.to_csv(rel_path, index=False)

    # Here the user sepcifiies number of channels. Exception shouldn't be thrown
    run_experiment(input_features, output_features, dataset=rel_path)

    del input_features[0]['preprocessing']['num_channels']

    # User now doesn't specify num channels. Should throw exception
    with pytest.raises(ValueError):
        run_experiment(input_features, output_features, dataset=rel_path)

    # Delete the temporary data created
    shutil.rmtree(image_dest_folder)


@pytest.mark.parametrize('encoder', ['wave', 'embed'])
def test_experiment_date(encoder, csv_filename):
    input_features = [date_feature()]
    output_features = [category_feature(vocab_size=2)]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    input_features[0]['encoder'] = encoder
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize('encoder', H3InputFeature.encoder_registry.keys())
def test_experiment_h3(encoder, csv_filename):
    input_features = [h3_feature()]
    output_features = [binary_feature()]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    input_features[0]['encoder'] = encoder
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_vector_feature_1(csv_filename):
    input_features = [vector_feature()]
    output_features = [binary_feature()]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_vector_feature_2(csv_filename):
    input_features = [vector_feature()]
    output_features = [vector_feature()]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_sampled_softmax(csv_filename):
    # Multiple inputs, Single category output
    input_features = [text_feature(vocab_size=10, min_len=1)]
    output_features = [category_feature(
        vocab_size=500,
        loss={'type': 'sampled_softmax_cross_entropy'}
    )]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename,
                             num_examples=10000)

    run_experiment(input_features, output_features, dataset=rel_path)


if __name__ == '__main__':
    """
    To run tests individually, run:
    ```python -m pytest tests/integration_tests/test_experiment.py::test_name```
    """
    pass



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import os
import platform
import shlex
import subprocess
import sys

import pytest
import tensorflow as tf

try:
    from horovod.common.util import nccl_built
except ImportError:
    HOROVOD_AVAILABLE = False
else:
    HOROVOD_AVAILABLE = True

from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import sequence_feature
from tests.integration_tests.utils import ENCODERS

# This script will run the actual test model training in parallel
TEST_SCRIPT = os.path.join(os.path.dirname(__file__), 'scripts',
                           'run_train_horovod.py')


def _nccl_available():
    if not HOROVOD_AVAILABLE:
        return False

    try:
        return nccl_built()
    except AttributeError:
        return False
    except RuntimeError:
        return False


def _run_horovod(csv_filename, **ludwig_kwargs):
    """Execute the training script across multiple workers in parallel."""
    input_features, output_features, rel_path = _prepare_data(csv_filename)
    cmdline = [
        'horovodrun',
        '-np', '2',
        sys.executable, TEST_SCRIPT,
        '--rel-path', rel_path,
        '--input-features', shlex.quote(json.dumps(input_features)),
        '--output-features', shlex.quote(json.dumps(output_features)),
        '--ludwig-kwargs', shlex.quote(json.dumps(ludwig_kwargs))
    ]
    exit_code = subprocess.call(' '.join(cmdline), shell=True,
                                env=os.environ.copy())
    assert exit_code == 0


def _prepare_data(csv_filename):
    # Single sequence input, single category output
    input_features = [sequence_feature(reduce_output='sum')]
    output_features = [category_feature(vocab_size=2, reduce_input='sum')]

    input_features[0]['encoder'] = ENCODERS[0]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    return input_features, output_features, rel_path


@pytest.mark.skipif(platform.system() == "Windows",
                    reason="Horovod is not supported on Windows")
@pytest.mark.distributed
def test_horovod_implicit(csv_filename):
    """Test Horovod running without `backend='horovod'`."""
    _run_horovod(csv_filename)


@pytest.mark.skipif(platform.system() == "Windows",
                    reason="Horovod is not supported on Windows")
@pytest.mark.skipif(not _nccl_available(),
                    reason="test requires Horovod with NCCL support")
@pytest.mark.skipif(not tf.test.is_gpu_available(cuda_only=True),
                    reason="test requires multi-GPU machine")
@pytest.mark.distributed
def test_horovod_gpu_memory_limit(csv_filename):
    """Test Horovod with explicit GPU memory limit set."""
    ludwig_kwargs = dict(
        gpu_memory_limit=128
    )
    _run_horovod(csv_filename, **ludwig_kwargs)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import multiprocessing
import os

import dask
import dask.array as da
import dask.dataframe as dd
from dask.diagnostics import ProgressBar

from ludwig.constants import NAME, PROC_COLUMN
from ludwig.data.dataset.parquet import ParquetDataset
from ludwig.data.dataset.partitioned import PartitionedDataset
from ludwig.data.dataframe.base import DataFrameEngine
from ludwig.utils.data_utils import DATA_PROCESSED_CACHE_DIR, DATASET_SPLIT_URL, DATA_TRAIN_HDF5_FP
from ludwig.utils.fs_utils import makedirs, to_url
from ludwig.utils.misc_utils import get_combined_features, get_proc_features
from ludwig.data.dataframe.dask_df_utils import dask_to_tfrecords


TMP_COLUMN = '__TMP_COLUMN__'


def set_scheduler(scheduler):
    dask.config.set(scheduler=scheduler)


class DaskEngine(DataFrameEngine):
    def __init__(self, parallelism=None, persist=False, **kwargs):
        self._parallelism = parallelism or multiprocessing.cpu_count()
        self._persist = persist

    def set_parallelism(self, parallelism):
        self._parallelism = parallelism

    def empty_df_like(self, df):
        # Our goal is to preserve the index of the input dataframe but to drop
        # all its columns. Because to_frame() creates a column from the index,
        # we need to drop it immediately following creation.
        return df.index.to_frame(name=TMP_COLUMN).drop(columns=[TMP_COLUMN])

    def parallelize(self, data):
        return data.repartition(self.parallelism)

    def persist(self, data):
        return data.persist() if self._persist else data

    def compute(self, data):
        return data.compute()

    def from_pandas(self, df):
        return dd.from_pandas(df, npartitions=self.parallelism)

    def map_objects(self, series, map_fn, meta=None):
        meta = meta or ('data', 'object')
        return series.map(map_fn, meta=meta)

    def apply_objects(self, df, apply_fn, meta=None):
        meta = meta or ('data', 'object')
        return df.apply(apply_fn, axis=1, meta=meta)

    def reduce_objects(self, series, reduce_fn):
        return series.reduction(reduce_fn, aggregate=reduce_fn, meta=('data', 'object')).compute()[0]

    def to_parquet(self, df, path):
        with ProgressBar():
            df.to_parquet(
                path,
                engine='pyarrow',
                write_index=False,
                schema='infer',
            )

    def to_tfrecord(self, df, path):
        """Implementations of data frame to tfrecords."""
        with ProgressBar():
            dask_to_tfrecords(
                df,
                path,
                compression_type="GZIP",
                compression_level=9)

    @property
    def array_lib(self):
        return da

    @property
    def df_lib(self):
        return dd

    @property
    def parallelism(self):
        return self._parallelism

    @property
    def partitioned(self):
        return True



import tensorflow as tf


# taken from https://github.com/google-research/google-research/blob/master/tabnet/tabnet_model.py
def glu(x, n_units=None):
    """Generalized linear unit nonlinear activation.

    Expects 2*n_units-dimensional input.
    Half of it is used to determine the gating of the GLU activation
    and the other half is used as an input to GLU,
    """
    return x[:, :n_units] * tf.nn.sigmoid(x[:, n_units:])


def gelu(features, approximate=False):
    if approximate:
        coeff = tf.cast(0.044715, features.dtype)
        return 0.5 * features * (
                1.0 + tf.math.tanh(0.7978845608028654 *
                                   (features + coeff * tf.math.pow(features,
                                                                   3))))
    else:
        return 0.5 * features * (1.0 + tf.math.erf(
            features / tf.cast(1.4142135623730951, features.dtype)))
